/*
 *
 * @(#)BufferCanvas.java 0.1 97/08/11
 *
 * Copyright (c) 1997 Ginger Booth.  All Rights Reserved.
 *
 * @history: original
 * @author Ginger Booth
 * @version 0.1, August 1997
 */
package courseware.jfcdep;

/** A double-buffered canvas.  Draw offscreen.
*/
import java.awt.*;
import courseware.util.*;

public class BufferCanvas extends Component {
    protected Image offscreen;  // draw offscreen to cut the flicker
    protected Graphics offg;    // the graphics context of the offscreen image
                                // min=preferred size, requested
    protected Dimension prefSize = new Dimension();
                                // am is the current reality
    private Dimension amSize   = new Dimension();

    // not actually used in Java1.1 version
    public final static float DEFAULT_CIRCLE_THICKNESS = 1.8f;
    public final static float DEFAULT_LINE_THICKNESS = 1.8f;

    public BufferCanvas(int prefsizex, int prefsizey) {
        prefSize      = new Dimension(prefsizex, prefsizey);
        amSize        = new Dimension(-1, -1);
    }

    public BufferCanvas(Dimension d) {
        prefSize      = new Dimension(d.width, d.height);
        amSize        = new Dimension(-1, -1);
    }
    public BufferCanvas() {this(1,1);}

    // force all the graphics to start over
    public void reset() {
        if (offscreen != null) {
            offscreen.flush();
            offscreen = null;
        }
        amSize.width = -1;
    }

    /** Ensure we have an offscreen image to draw on, and that the scaling
     *  matches its dimensions.
     */
    synchronized public Graphics getOffscreen() {
        Dimension s = getSize();
        if (s.width < 1) s.width = 1;
        if (s.height < 1) s.height = 1;
        if ( (s.width  != amSize.width) ||
             (s.height != amSize.height)) {

            if (offscreen != null) {
                offscreen.flush();
            }

            offscreen     = createImage(s.width, s.height);
            if (offscreen != null) {
              amSize.width  = s.width;
              amSize.height = s.height;
              offg          = offscreen.getGraphics();
            }
        }
        return offg;
    }

    /** Ensure we have a NEW offscreen image to draw on.
     */
    protected Graphics newOffscreen() {
        Dimension s = getSize();

        if (offscreen != null)
            offscreen.flush();

        amSize.width  = s.width;
        amSize.height = s.height;
        offscreen     = createImage(amSize.width, amSize.height);
        offg          = offscreen.getGraphics();

        return offg;
    }

    // put offscreen graphics onscreen
    public void update(Graphics g) { paint(g); }

    // put offscreen graphics onscreen
    public void paint(Graphics g) {
        getOffscreen();
        if (offscreen != null)
           g.drawImage(offscreen, 0, 0, this);
    }
    // put offscreen graphics onscreen
    public void onscreen() {
        if (offscreen != null)
           getGraphics().drawImage(offscreen, 0, 0, this);
    }

    synchronized public void setSize(int w, int h)   {
        Dimension d = new Dimension(w, h);
        this.setSize(d);
    }

    synchronized public void setSize(Dimension d)   {
        prefSize = d;
        super.setSize(d);
        getOffscreen();
    }

    synchronized public void setBounds(int x, int y, int width, int height)   {
        Dimension d = new Dimension(width, height);
        prefSize = d;
        super.setBounds(x,y,width,height);
        getOffscreen();
    }

    public Dimension getPreferredSize() { return prefSize; }
    public void setPreferredSize(Dimension d) {prefSize = d; amSize=new Dimension(-1, -1);}
    public Dimension getMinimumSize()   { return prefSize; }
    public Dimension getCurrentSize()   { return amSize; }

    /** Release the graphics allocation. */
    public void finalize() {
        if (offscreen != null)
            offscreen.flush();
        offscreen = null;
        try {
            super.finalize();
        } catch (Throwable t) {};
    }
    /** calls drawOval. JFC version can use Graphics2D features for antialias */
    public void drawCircle(int ulx, int uly,
                           int diameter, Color c)
    {
      Graphics g = getOffscreen();
      g.setColor(c);
      g.drawOval(ulx, uly, diameter, diameter);
      g.drawOval(ulx-1, uly-1, diameter+2, diameter+2);
    }

    /** calls drawOval. JFC version can use Graphics2D features for antialias */
    public void drawCircle(float ulx, float uly,
                           float diameter, float thickness, Color c)
    {
      Graphics g = getOffscreen();
      g.setColor(c);
      int x = Math.round(ulx);
      int y = Math.round(uly);
      int diam = Math.round(diameter);
      g.drawOval(x, y, diam, diam);
      g.drawOval(x-1, y-1, diam+2, diam+2);
    }
    /** calls fillOval. JFC version can use Graphics2D features for antialias */
    public void fillCircle(float ulx, float uly,
                           float diameter, Color c)
    {
      Graphics g = getOffscreen();
      g.setColor(c);
      int x = Math.round(ulx);
      int y = Math.round(uly);
      int diam = Math.round(diameter);
      g.fillOval(x, y, diam, diam);
    }
    /** calls fillOval. JFC version can use Graphics2D features for antialias */
    public void fillCircle(int ulx, int uly,
                           int diameter, Color c)
    {
      Graphics g = getOffscreen();
      g.setColor(c);
      g.fillOval(ulx, uly, diameter, diameter);
    }

    /** calls drawString. JFC version can use Graphics2D features for antialias */
    public void drawString(Color color, String s, int x, int y, boolean antialias) {
      drawString(color, s, x, y);
    }
    /** calls drawString. JFC version can use Graphics2D features for antialias */
    public void drawString(Color color, String s, int x, int y) {
      Graphics g = getOffscreen();
      g.setColor(color);
      g.drawString(s, x, y);
    }
    /** calls drawLine. JFC version can use Graphics2D features for antialias */
    public void drawLine(Color c, int x0, int y0, int x1, int y1, float thickness) {
      Graphics g = getOffscreen();
      g.setColor(c);
      g.drawLine(x0, y0, x1, y1);
    }
    /** calls drawPolyline. JFC version can use Graphics2D features for antialias */
    public void drawLine(Color c, int[] x, int[] y, int npts, float thickness) {
      Graphics g = getOffscreen();
      g.setColor(c);
      g.drawPolyline(x, y, npts);
    }
    static public void drawLine(Graphics g,
      Color c, int[] x, int[] y, int npts, float thickness)
    {
      g.setColor(c);
      g.drawPolyline(x, y, npts);
    }


} /* BufferCanvas */


