//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990331, vfb, original Java
 * Translated from the Swarm Objective-C Gecko source, version 0.51
 */
/** LandRoot is the Quad rooting all the resource-producing Quads.
 *  It does bookkeeping and divvying for its subordinate Quads that
 *  actually produce. For Gecko's sunwater production, this is presently
 *  done by ResourceQuad's. For BacSim, DiffuseQuad's.
 *
 *  Part of the divvying done by LandRoot is all the picking chicanery that
 *  makes plants space away from each other. In other words, LandRoot actively
 *  feeds sessile agents by dividing up the spoils between them, and setting
 *  the vectors that in net make plants grow toward the sun. See feedCritter.
 *
 *  The originating version of Gecko had hooks but no implementation to
 *  allow conversion of taxes (metabolic byproducts) back into produced
 *  resources. Hooks left out of translation.
 */


package gecko;
import gecko.zoo.*;

public class LandRoot extends Quad {
  private ResourceList startres;	// bookkeeping--overall at end of stepResources
  private ResourceList reservoir; // bookkeeping--overall now
  private ResourceList taxTake;	  // bookkeeping--overall

  // avoid allocating these scratchpads in eatCompete
  private static McrCircle feedwin = new McrCircle();
  private static McrCircle eatersquare = new McrCircle();

  public LandRoot() {
    super();
    setVisible(true);
    setType(QUAD_FEED);
    startres  = new ResourceList();
    reservoir = new ResourceList();
    taxTake   = new ResourceList();
  }

  /** feedCritter takes an agent, eater, and feeds it a given percent of the
   *  available resource over the given area. The returned value is how much
   *  of the resource that turned out to be. But, note that the eater and
   *  the eaten have already be credited/debited when feedCritter returns.
   */
  public double feedCritter(Critter eater, int resource, double percent, Mcr area) {
    Quad qsite;
    double got = 0.0;

    qsite = pickWin(area, this, true);

    while (qsite != null) {
      if ((qsite instanceof ResourceQuad) && ((ResourceQuad)qsite).produces(resource))
        got += ((ResourceQuad)qsite).feedCritter(eater, percent, area);
      qsite = qsite.pickWin(area, this, true);
    }
    reservoir.list[resource] -= got;

    return got;
  }

  /**
    // eatCompete: Feed the plantRoot tree of agents.  Tricky, in
    // that overlapping agents must share resources, as feed is an area
    // effect.  To do this:
    // 	a. Use each agent in turn as a pick window (eater).  Find its area.
    //  b. For each agent (competitor) picked with that window, subtract
    //     one half the intersecting area from the eater.
    //  c. When done picking competitors, we have the effective area of
    //     the eater.  Express this as a percent of the original.
    //  d. Send the FeedQuad the true eater mcr and %area, and request
    //     how much the eater gets fed.
    // Note:  this is not -correct-, in that if more than two agents
    // overlap, the effective area decreases faster than it should.  E.g.,
    // if three agents overlap, they should each get 1/3 of the resources
    // from the 3-way overlap, but in this scheme, they'll each get 1/4.
    // As of this writing, I weakly defend this as the price of competition,
    // and more strongly defend it as the most compute-effective way I could
    // think of to localize errors in an intuitively clear fashion.

    // This routine resets agent scratchpads and visibility across the tree,
    // as well as feeding the agents and debiting the landscape.

    // Note that competition is calculated separately for each resource.
    */
  private void eatCompete(Quad root) {
    // eater and competitor can be considered two parsing pointers on tree
    int res;
    Quad equad, cquad;
    PassiveCritter eater;
    PassiveCritter competitor;
    double eaterarea;
    double eaterarea1;	// unmodified eater area.
    double contestedarea;
    double wx, wy;	// for calculating wannago direction
    int ecount;
    
    // clip feeding area and tree area to each other
    if (intersectMcr(root, feedwin) == MCR_DISJOINT)
      return;

    // First, pick everyone in the area and set their scratchpads to zero.
    // All picking series go by same root.
    equad = root.pickWin(feedwin, root, false);
    if (equad == null)
      return;

    while (equad != null) {
      if (equad instanceof PassiveCritter)
        ((PassiveCritter) equad).setScratchpad(0.0);
      equad = equad.pickWin(feedwin, root, false);
    }

    // now back to first eater
    ecount = 0;
    equad = root;
    while ( (equad = equad.pickWin(feedwin, root, false)) != null) {
      if (!(equad instanceof PassiveCritter))
        continue;
      eater = (PassiveCritter) equad;
      ecount++;

      // set this guy invisible to the next pick operation
      eater.setVisible(false);

      // set eater's equivalent square now.
      eatersquare.circleToSquare((Critter) eater);

      // subtract off results of prior picks
      eaterarea = eaterarea1 = eatersquare.getArea();
      eaterarea -= eater.getScratchpad();

      // consider each competitor
      cquad = root.pickWin(eater, root, false);
      while (cquad != null) {
        if ( (cquad instanceof PassiveCritter) &&
             eater.competesWith( (PassiveCritter) cquad) ) {
          competitor = (PassiveCritter) cquad;
          contestedarea = eatersquare.crossCircles(eater, competitor);
          // NOTE:  because we're calculating debits on competitors as
          // well as debits on the eater, we must go on to the bitter end
          // even if eater's area becomes negative.  As it may because of
          // the error source noted above.

          contestedarea /= 2.0;
          eaterarea -= contestedarea;
          competitor.setScratchpad(competitor.getScratchpad() + contestedarea);
// System.err.println("t=" + Globals.getCurrentTime() +
// ", contestedarea=" + (contestedarea/eaterarea1) + " percent of eater");

          // now, let's set wannas -- away from competitor
          wx = competitor.getX() - eater.getX();
          wy = competitor.getY() - eater.getY();
          eater.addDirection(-wx, -wy, contestedarea);
          competitor.addDirection(wx, wy, contestedarea);
        }
        cquad = cquad.pickWin(eatersquare, root, false);
      }

      // done competing, let's eat.
      if (eaterarea > 0.0) {
        for (int r=1; r <=Globals.getResources().getNumResources(); r++)  {
          if (eater.eats(r)) {
            feedCritter(eater, r, (eaterarea/eaterarea1), eatersquare);
// System.err.println("feed " + (eaterarea/eaterarea1) + "% over " + eatersquare
// + ", eater r=" + eater.getRadius());
          }
        }
      }
    }

    // Now for cleanup.  Must set our invisibles back to visible.
    equad = root.pickWin(feedwin, root, true);
    while (equad != null) {
      if (equad instanceof PassiveCritter)
        ((PassiveCritter) equad).setVisible(true);
      equad = equad.pickWin(feedwin, root, true);
    }
  }

  /** If you need to talk to the resource quad directly...returns first it
   *  finds that produces this resource over that area.  If trying to find for
   *  an agent, pass self (self being an mcr) as the area (or its equiv-area square).
   */
  public ResourceQuad firstQuadProducing(int res, Mcr overMcr) {
    Quad site;
    site = pickWin(overMcr, this, true);
    while (site != null) {
      if ((site instanceof ResourceQuad) && (((ResourceQuad)site).produces(res)) )
        return (ResourceQuad) site;
      site = site.pickWin(overMcr, this, true);
    }
    return null;
  }

  public ResourceQuad nextQuadProducing(int res, Mcr overMcr, Quad resumefrom) {
    Quad site;
    
    site = resumefrom.pickWin(overMcr, this, true);

    while (site != null) {
      if ((site instanceof ResourceQuad) &&  (((ResourceQuad)site).produces(res)) )
          return (ResourceQuad) site;
      site = site.pickWin(overMcr, this, true);
    }
    return null;
  }

  /** Should be called with subtree on which to find passiveCritters.*/
  public void feedTree(Quad root) {eatCompete(root);}
  public void feedTreeOn(Quad root, int roundth) {
    if (Globals.getCurrentTime() % roundth == 0)
      feedTree(root);
  }

  /** For now, we just tally up. The FIRST suitable ResourceQuad gets all.
   *  Later, maybe convert.
   */
  public void payTax(double x, double y, int res, double amount) {
    Quad site;
    ResourceQuad rsite;
    taxTake.list[res] += amount;
    
    // any takers? all goes to first volunteer
    site = pickWin(this, this, true);

    while (site != null) {
      if ( (site instanceof ResourceQuad) ) {
        rsite = (ResourceQuad) site;
        if ( rsite.produces(res) && rsite.isInsideOrOn(x,y) ) {
          rsite.addResource(amount, x, y);
          return;
        }
      }
      site = site.pickWin(this, this, true);
    }
  }
  
  public void payTax(double x, double y, ResourceList list) {
    for (int i=0; i < list.list.length; i++)
      if (list.list[i] > 0.0)
        payTax(x, y, i, list.list[i]);
  }

  public double getTaxesPaid(int res) { return taxTake.list[res]; }

  public ResourceList getTaxesPaid() {return (ResourceList) taxTake.clone(); }

  public void zeroTaxes() {taxTake.zero();}

  public void stepResourcesOn(int roundth) {
    if ((Globals.getCurrentTime() % roundth) == 0)
      stepResources();
  }

  /** call on landscape collector */
  public void stepResources() {
    Quad asite;
    ResourceQuad rsite;

    startres.zero();
    zeroTaxes();

    asite = pickWin(this, this, true);
    while (asite != null) {
      if (asite instanceof ResourceQuad) {
        rsite = (ResourceQuad) asite;
        rsite.stepResources();
        startres.list[rsite.getResource()] += rsite.getStartReservoir();
      }
      asite = asite.pickWin(this, this, true);
    }

    reservoir.zero();
    reservoir.add(startres);
  }

  /** call on landscape collector
   *  NOTE: There's a big difference between stepResources, which prepares
   *  for the new round, and step(), which does stats collection at the end
   *  of the round. The more interesting landscape stats have to do with the
   *  results of the round, how much fed out and how much tax paid.
   */
  public void step() {
    Quad asite;
    asite = pickWin(this, this, true);
    while (asite != null) {
      if (asite instanceof ResourceQuad) {
        ((ResourceQuad) asite).step();
      }
      asite = asite.pickWin(this, this, true);
    }
  }

  /** whole tree, current feed on offer. */
  public ResourceList curFeed () { return (ResourceList) reservoir.clone(); }

  /** startres - reservoir, whole tree */
  public ResourceList fedOut() {
    ResourceList totals = (ResourceList) reservoir.clone();
    totals.subtract(startres);
    return totals;
  }

  /** fedOut expressed as percentage */
  public ResourceList fedOutPercent() {
    ResourceList totals = (ResourceList) reservoir.clone();

    for (int i=0; i < totals.list.length; i++) {
      if (startres.list[i] == 0.0)
        totals.list[i] = 0.0;
      else
        totals.list[i] /= startres.list[i];
    }
    return totals;
  }

  public String toString() {
    return "LandRoot[" + super.toString();
  }
}
