//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator

/** Thread-runnable gizmo to control simulation. Lifted from Indeco project. */

package gecko;

import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.util.*;
import java.awt.*;
import stats.*;
import gecko.camera.*;
import gecko.bacsim.BacGlobals;

import gecko.woods.*; /* for debug */

public class Simulator implements Runnable {
  static private double maxx = 10;
  static private double maxy = 10;
  static private long stopTime = 10;
  static private long pauseTime = -1; // not end of run, but end of step10 or whatever.
  static private int showEvery = 1;
  static private int statsEvery = 1;
  static private Vector species = new Vector(1,1); // Vector of Species
  static private Container driver;
  // basenap is milliseconds
  static private long basenap = 40; // was int 20, before that 4...we sleep rather than yield, seems to work better....
  // current state of play....
  static private Thread runningThread = null;
  static private boolean interruptFlag = false;
  static private boolean paramsLoadingFlag = false;   // skip restart from params file
  static private boolean inStepFlag = false; // step in progress

  static private Steppable camera = null; // could generalize, but let's keep it simple
  static private Vector miscSteppable = null; // generic Steppable expansion

  public void setStopTime(int n) {stopTime = (long) n;}
  public void setSeason(int n) {Globals.setSeasonCycle(n);}
  public void setShowEvery(int n) {showEvery = n;}
  public void setStatsEvery(int n) {statsEvery = n; Globals.setStatCycle(n);}
  public void setSortEvery(int n) {Globals.setSortCycle(n);}
  public void setMaxX(double d) {maxx = d;}
  public void setMaxY(double d) {maxy = d;}
  public void setMaxXY(double d) {maxx = maxy = d;}
  public void setDriver(Container whoToRepaint) {driver = whoToRepaint;}
  /** driver is who to repaint*/
  static public Container getDriver() {return driver;}

  public Simulator() {
    setStatsEvery(statsEvery); // let globals start in sync....
  }

  public void clear() {
    interruptFlag = false;
    pauseTime = -1;
    miscSteppable = null;
    Species.removeAllSpecies(); // population back to 0, etc.
    Site.removeAllSites(); // population back to 0, etc.
    if (Globals.getUI().isGUI())
      if (Globals.getWorldDisplay() != null)
        Globals.getWorldDisplay().clearLegend();
    Globals.getStats().clear(); // clear resets plots, want all plots gone.
    Globals.setCurrentTime(0);
    MemCleaner.nap();
    Globals.getStats().init(); // clear resets plots, want all plots gone.
    makeWorld();
    MemCleaner.nap();
  }

  public void close() {
    QuadDisplay qd = Globals.getWorldDisplay();
    qd.setVisible(false);
    Globals.setWorldDisplay(null);
    qd.finalize();
  }

  public Thread getThread() { return runningThread; }

/** Woods double-check on population and basal area */
private void checkTrees() {
Quad plantroot = Globals.getPlantRoot();
Quad quad;
Tree tree;
double totalBA = 0;
double pop = 0;
double light = 0;
int[] buckets = new int[8];
int[] lbuckets = new int[10];
double dbucket = .25;

quad = plantroot.pickWin(plantroot,plantroot);
while (quad != null) {
  if (quad instanceof Tree) {
    tree = (Tree) quad;
    pop += 1.0;
    light += tree.getLight();
    int i = Math.round(Math.round(tree.getLight() * 10));
    if (i==10) i=9;
    lbuckets[i]++;
    totalBA += tree.getBiomass();
    i = Math.round(Math.round(Math.floor(tree.getBiomass() / dbucket)));
    if (i>7) i = 7;
    buckets[i]++;
  }
  quad = quad.pickWin(plantroot,plantroot);
}
System.err.println("Time " + Globals.getCurrentTime() + ", pop=" + pop + ", BA=" + totalBA
+ ", avgBA=" + (totalBA/pop) + ", avg light=" + (light/pop) );
System.err.print("-- BA distrib : ");
for (int i=0; i<8; i++)
System.err.print( TruncVal.truncVal( buckets[i]/pop ) + "  ");
System.err.print("-- light distrib : ");
for (int i=0; i<10; i++)
System.err.print( TruncVal.truncVal( lbuckets[i]/pop ) + "  ");

System.err.println();
}

  synchronized public void step() {
    inStepFlag = true;
    int maxnaps = 100;
// MemCleaner.setVerbose(true);

    // these should happen in my thread....
    Species.stepAll();
    Site.stepAll();
// System.err.println("\n\nstep " + Globals.getCurrentTime() + " after Species/site stepAll");
// Globals.getWorldRoot().printTree();

/*
    if (Species.getLastBirthday() == Globals.getCurrentTime()) {
      Globals.getWorldRoot().rootQuad();
      MemCleaner.cleanup();
    }
*/

    // note in oldGecko, deathdayCheck and onSeason were handled from
    // GeckoModelSwarm -> Species/Site. They've been demoted here into
    // the beginning of the critter step.

    Globals.getLandRoot().stepResources();
    Globals.getLandRoot().feedTree(Globals.getPlantRoot());
    Globals.getPlantRoot().step();
    Globals.getMobileRoot().step();
    if (miscSteppable != null) {
      for (int i=0; i<miscSteppable.size(); i++)
        ((Steppable) miscSteppable.elementAt(i)).step();
    }

    // some Woods analysis debug
    // if ((Globals.getCurrentTime() % 50) == 0)
    //   checkTrees();

    // moved sorting before drawing so that we can't remake the tree
    // while a redraw is going on .... Netscape doesn't respond well to
    // our means of saying "do the repaints NOW"....
    if ( (Globals.getCurrentTime() % Globals.getSortCycle()) == 0 )  {
      Globals.getWorldRoot().rootQuad();
      MemCleaner.cleanup();
    }

    if ( (Globals.getCurrentTime() % showEvery) == 0 ) {
      if (camera != null)
        camera.step();

      if (Globals.getUI().isGUI()) {
        synchronized (Globals.getWorldDisplay()) {
          Globals.getWorldDisplay().repaint();
          try {
            Thread.currentThread().sleep(basenap);
          } catch (Exception e) {}
//          runningThread.yield(); // make sure it STARTS now, nap loop only works if started
        }
        int napno = 1;
        do {
          MemCleaner.nap(); // now make sure it finishes.....
          napno++;
        } while (Globals.getWorldDisplay().isDrawing() && (napno < maxnaps) );
// if (napno == maxnaps)
// System.err.println("Simulator step, napno=" + napno + ", maxnaps=" + maxnaps
// + ", time=" + Globals.getCurrentTime());
      }
    }
    // repaints never happen in my thread....
    if ( (Globals.getCurrentTime() % statsEvery) == 0 ) {
      // landscape step is a book-keeping step
      Globals.getLandRoot().step();
      Species.stepAllStats();
      Globals.getStats().step();
      // have stat data already, NOW redraw it
      Globals.getUI().step();
      int napno = 0;
      do {
        MemCleaner.nap(); // now make sure it finishes.....
        napno++;
      } while (Globals.getStats().isStepping() && (napno < maxnaps) );
    }

    Globals.incrementTime();
    inStepFlag = false;
  }

  public void run() {
    runningThread = Thread.currentThread();
    long now;
    // give AWT right of way....
    while ( true ) {
      // let's untangle the termination conditions....
      now = Globals.getCurrentTime();
      if (interruptFlag)
        break;
      if (now == pauseTime)
        break;

      // otherwise, do a step
      step();
      now = Globals.getCurrentTime();
      if ( (now == (stopTime + 1)) && (now > pauseTime) )
        break;
    }
    interruptFlag = false; // have interrupted, that's all it's for.
  }

  static public void pause() {
    interruptFlag = true;
    int maxhang = 30;
    int naps = 0;
    while (isStillStepping() && (naps < maxhang)) {
      MemCleaner.nap();
      naps++;
    }
  }

  synchronized public void simulateMore() {
    start();
  }
  synchronized public void simulateContinue() {
    pauseTime = -1;
    simulateMore();
  }

  public void simulateMore(int steps) {
    pauseTime = Globals.getCurrentTime() + steps;
    simulateMore();
  }
  public void simulateMore1() {simulateMore(1);}
  public void simulateMore10() {simulateMore(10);}

  /* for a debug button, created by GeckoDriverCommon */
  public void dumpTree() {
    Globals.getWorldRoot().printTree();
  }

  public void postParams() {
    Quad mobileroot = Globals.getMobileRoot();
    if ((mobileroot != null) && (mobileroot.getParent() == Globals.getWorldRoot())) {
      mobileroot.deleteBelow();
      mobileroot.remove();
    }
    mobileroot = new Quad();
    mobileroot.setType(Quad.QUAD_MOBILE);
    Globals.getWorldRoot().addChild(mobileroot);
    Globals.setMobileRoot(mobileroot);

    Quad plantroot = Globals.getPlantRoot();
    if ((plantroot != null) && (plantroot.getParent() == Globals.getWorldRoot())) {
      plantroot.deleteBelow();
      plantroot.remove();
    }
    plantroot = new Quad();
    plantroot.setType(Quad.QUAD_SESSILE);
    Globals.getWorldRoot().addChild(plantroot);
    Globals.setPlantRoot(plantroot);

    LandRoot land = Globals.getLandRoot();
    if ((land != null) && (land.getParent() == Globals.getWorldRoot())) {
      land.deleteBelow();
      land.remove();
    }
    land = new LandRoot();
    Globals.getWorldRoot().addChild(land);
    Globals.setLandRoot(land);

    Globals.setWorldMcr(0,0,maxx,maxy);
    Site.hangAllSites();
    Globals.setTaxCollector(land);

    if ( Globals.getGeckoDriver().isBacsim()) {
      BacGlobals.init();
    }

    if (Globals.getUI().isGUI()) {
      // Globals.worldDisplay.displayQuads = true;
      Globals.getWorldDisplay().centerPicture(Globals.getWorldMcr());
      driver.repaint();
    }
    MemCleaner.nap();
  }

  public void addSpecies(Species si) {
    // re-register this thing with Species.
    si.register();
  }
  public void addSpecies(CompositeSpecies si) {
    // re-register this thing with Species.
    Species.register(si);
  }
  public void addSite(Site si) {
    // re-register this thing with Site.
    si.register();
  }

  /** The miscSteppable is stepped each step, after all the non-misc
   *  sites & species & agents, but before the stats.
   */
  public void addMiscSteppable(Steppable misc) {
    if (miscSteppable==null)
      miscSteppable = new Vector(1,1);
    miscSteppable.addElement(misc);
  }

  /** The camera is stepped on displaycycle, regardless of
   *  whether this is a GUI run. For the nonce, only Simulator
   *  knows about the camera.
   */
  public void addPOVCamera(String includeMacrosFile) {
    camera = new POVCamera(includeMacrosFile);
  }

  /** By default, the camera takes everything in the World Mcr. This limits the
   *  view. Agents crossing out of the field simply disappear.
   */
  public void setPOVCameraField(double xmin, double ymin, double xmax, double ymax) {
    if ((camera != null) && (camera instanceof POVCamera))
      ((POVCamera) camera).setField(xmin, ymin, xmax, ymax);
  }
  public void turnCameraOff() {camera=null;}

  synchronized public void makeWorld() {
    inStepFlag = true;
    Globals.setWorldRoot(new Quad());
    Globals.setWorldMcr(0,0,maxx,maxy);

    if (Globals.getUI().isGUI()) {
      if ( Globals.getWorldDisplay() == null) {
        Globals.setWorldDisplay(new QuadDisplay(Globals.getWorldRoot()));
        Globals.getWorldDisplay().setVisible(true);
        driver.add(Globals.getWorldDisplay(), ModuleLayout.BIG_STRING);
        // driver.setVisible(true) was here
        ModuleLayout ml = (ModuleLayout) driver.getLayout();
        ml.setConstraint(Globals.getWorldDisplay(), 0, ModuleLayout.BIG);
        driver.setVisible(true);
      } else {
        Globals.getWorldDisplay().setVisible(false);
        Globals.getWorldDisplay().setRoot(Globals.getWorldRoot());
        Globals.getWorldDisplay().setVisible(true);
      }
      driver.repaint();
      MemCleaner.cleanup();
    }
    inStepFlag = false;
    Species.resetStatsAll(); // population back to 0, etc.
    Site.resetStatsAll();    // population back to 0, etc.
    Globals.getUI().reset(); // reuse same stats, reset to time 0.
    Globals.getStats().reset(); // reuse same stats, reset to time 0.
  }

  public void restart() {
    if (paramsLoadingFlag)
      return;
    if ((runningThread != null) && (runningThread.isAlive()))
      runningThread.stop();
    Globals.setCurrentTime(0);
    makeWorld();
    postParams();
    start();
  }

  public void setParamsLoading(boolean b) {paramsLoadingFlag = b;}

  static public boolean isStillStepping() {return inStepFlag;}

  /** Don't proceed--kill thread. Should still be able to join(0) the
   *  thread (wait until finished) in batch multi-runs.
   */
  static public void crashStop() {
    if ( (runningThread != null) && (runningThread.isAlive()) )
      runningThread.stop();
    interruptFlag = true;
  }

  synchronized public void start() {
    inStepFlag = false; // just in case...
    MemCleaner.setNapTime(basenap); // hopefully enuf--it matters!

    if ((runningThread != null) && (runningThread.isAlive())) {
      pause();
      runningThread.setPriority(Thread.MAX_PRIORITY);
      MemCleaner.nap();
      runningThread.stop();
      MemCleaner.nap();
    }
    runningThread = new Thread(this);
    runningThread.setPriority(Thread.MIN_PRIORITY);
    runningThread.start();
  }
}