//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator

/** Collects the aspect of the Speciesable best dealt with for the Species
 *  as a whole--birthday spiel, statistics, name, color, etc.  On the
 *  birthday, progenitor is cloned and hung on the quad tree (if it is
 *  a Quad.)
 *  Although Species is a StatSource, that's for birth/death listeners.
 *  In practice, Gecko doesn't use those listeners. Though they might be
 *  useful if species could emerge or go extinct in the course of a run.
 *  What they're actually for in the stat package is for notifying things
 *  that refer to the StatSource, that the StatSource is going away or a
 *  new one needs to be made. For current Gecko needs, that doesn't happen,
 *  as Species are specified and eternal at run begin.
 */

package gecko;

import stats.*;
import java.util.*;
import java.awt.Color;
import gecko.statglue.*;
import gecko.zoo.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.lang.reflect.*;

public class Species extends StatSource implements Steppable, Biomassable {
  static private Vector allSpecies = new Vector(1,1);
  static private Vector composites = new Vector(1,1); // CompositeSpecies--not stepped.
  private Speciesable progenitor;
  private String name;
  private long birthday;
  private int initNum;
  private int origInitNum; // used for seasonal changes
  private Mcr initArea; // if not given, defaults to present world, evenly.
  private static long lastBirthday = -1; // last realized birthday of any species
  private Vector varies; // if any, list of VaryProperty to apply after birth
  // may or may not add progenitorVaries at some point.

  /** This property list is used to generate paramLists for popup editing.
   *  Note initArea is left off for now, as don't have the routines to set
   *  it nicely from a popup.
   */
  private static String[] props = { "name", "birthday", "initNumber"  };


  // this unknown bit of whatever is being held for the progenitor class
  // as static data for the species. Cannot use a static within the class, or
  // all species of that class (maybe even inheriting that class) would share
  // the same values. First use of this is "onSeason", but I can imagine
  // other uses. 
  private Object holdData = null;       // could be anything--just hold it

  // Statistics
  private double population = 0;
  private double biomass = 0;
  private PopBroker popBroker;
  private BiomassBroker biomassBroker;
  private BiomassTransfers bioTrans;
  private BiomassBucketBroker[] bioTransBrokers;
  private BiomassTransfers bioRate;
  private BiomassBucketBroker[] bioRateBrokers;
  private StageStatBuckets stages;
  private StageObserver stageObserver;
  private StageBucketBroker[] fertBrokers;
  private StageBucketBroker[] mortBrokers;
  private StageBucketBroker[] stageBrokers;
  private FragStatBuckets fragStats; // added, perhaps, by CompositeSpecies
  private RadiusMeanSigmaSkew radii;
  private IndexStatBroker[] radiusBrokers;

  /** Species are not registered with allSpecies when created. */
  public Species(String name, Speciesable progenitor, int bornWhen, int quantity) {
    super();
    setName(name);
    setPrototype(progenitor);
    progenitor.setSpecies(this);
    setBirthday(bornWhen);
    setInitNumber(quantity);
    population = 0;
    biomass = 0;
  }

  public void register() {
    if (!allSpecies.contains(this)) {
      allSpecies.addElement(this);
    }
    resetStats();
  }
  static public void register(CompositeSpecies cs) {
    if (!composites.contains(cs)) {
      composites.addElement(cs);
    }
  }
  static public Vector getAllSpecies() { return allSpecies; }
  static public Vector getAllCompositeSpecies() { return composites; }
  public StageObserver getStageObserver(){return stageObserver;}
  /** takes precedence over any supplied by progenitor. */
  public void setStageObserver(StageObserver so){stageObserver = so;}

  static public void removeAllSpecies(){
    allSpecies.removeAllElements();
    composites.removeAllElements();
  }

  public void addFragStats(FragStatBuckets stats) {
    fragStats = stats;
  }

  private void addParamList(ParamList addTo, GetPropertyNamesable propsFor) {
    ParamList pnew = ParamListBuilder.paramListFor(propsFor,
                            propsFor.getPropertyNames());
    for (int i=0; i<pnew.size(); i++) {
      addTo.add(pnew.getElement(i));
    }
  }

  /** Tell this parameter to vary according to sigma between instances born
   *  of this progenitor. The parameterName is, like, "radius". A JavaBean
   *  "property name". At this writing, only double parameters can be varied.
   *  That's not checked. One thing that is checked is that only one vary is
   *  permitted per property name. If there's already one, its sigma is reset.
   */
  public void vary(String parameterName, double sigma) {
    if (varies == null) {
      varies = new Vector(1,1); 
    } else {
      for (int i=0; i<varies.size(); i++) {
        VaryProperty avary = ((VaryProperty)varies.elementAt(i));
        if ( avary.nameEquals(parameterName) ) {
          avary.setSigma(sigma);
          return;
        }
      }
    }
    VaryProperty newVary = new VaryProperty(parameterName, progenitor.getClass(),
      sigma, Globals.getRandomSource());

    varies.addElement(newVary);
  }

  /** for use by ScaredyGrazer onSeason */
  public void resetVaryRadius() {
    // if not keeping radius stats, get out
    if (radii == null) return;
    // find a varyRadius or get out
    if (varies == null) return;
    VaryProperty theVary = null;
    for (int i=0; i<varies.size(); i++) {
      VaryProperty avary = ((VaryProperty)varies.elementAt(i));
      if ( avary.nameEquals("radius") ) {
        theVary = avary;
      }
    }
    if (theVary == null) return;

    // ok, we do have a varyRadius to reset, and new values, and always progenitor
    theVary.setSigma(radii.getStat(MeanSigmaSkew.SIGMA));
    QuadCircleSpecies prog = (QuadCircleSpecies) progenitor;
    prog.setRadius(radii.getStat(MeanSigmaSkew.MEAN));
  }

  /** return ParamList suitable for popup purposes, for Species and its
   *  progenitor.
   */
  public ParamList getParamList() {
    ParamList pSpecies = ParamListBuilder.paramListFor(this, props);
    if (progenitor instanceof GetPropertyNamesable)
      addParamList(pSpecies, (GetPropertyNamesable) progenitor);
    // varies need to be last, since they work on current value=mean
    if (varies != null) {
      for (int i=0; i<varies.size(); i++) {
        pSpecies.add( ((VaryProperty)varies.elementAt(i)).makeParam());
      }
    }
    return pSpecies;
  }

  private ExecParam makeEP(String observeWhat) {
    ExecParam ep;
    try {
      Method observeM = this.getClass().getMethod("observe" + observeWhat,
        Param.NO_ARGS_SIGNATURE);
      ep = new ExecParam(observeWhat, Param.NO_ARGS, this, observeM);
    } catch (Exception e) {
      System.err.println("Couldn't observe " + name + " statistic " + observeWhat);
      return null;
    }
    return ep;
  }

  /** Build list of ExecParam's to turn on stats for one species. */
  public ParamList getStatParamList() {
    ParamList answer = new ParamList();
    answer.setName(name + " Stats");
    answer.add(makeEP("Population"));
    answer.add(makeEP("Biomass"));
    answer.add(makeEP("Stages"));
    answer.add(makeEP("Fertility"));
    answer.add(makeEP("Mortality"));
    answer.add(makeEP("BiomassTransfers"));
    answer.add(makeEP("BiomassRates"));
    answer.add(makeEP("BiomassGrowth"));
    answer.add(makeEP("BiomassTheft"));
    answer.add(makeEP("Radius"));
    return answer;
  }

  public void setHoldData(Object unknown) {holdData = unknown;}
  /** could certainly be null */
  public Object getHoldData() { return holdData;}

  static public Species find(String name) {
    Species s;
    for (int i=0; i< allSpecies.size(); i++) {
      s = (Species) allSpecies.elementAt(i);
      if (s.getName().equals(name))
        return s;
    }
    return null;
  }

  static public void resetStatsAll() {
    Species s;
    for (int i=0; i< allSpecies.size(); i++) {
      s = (Species) allSpecies.elementAt(i);
      s.resetStats();
    }
  }

  public void resetStatsRound() {
    biomass = 0;
  }

  public void resetStats() {
    population = 0;
    holdData = null;
    initNum = origInitNum;
    if (fragStats != null)
      fragStats.reinit();
    resetStatsRound();
  }

  /** formalized kludge--returns true once per run. Call
   *  this the first time -before- setting holdData.
   */
  public boolean isNewRun() {
    if (holdData==null) {
      holdData = new String("runMarker");
      return true;
    } else {
      return false;
    }
  }

  public void setBirthday(int bornwhen) {
    birthday=bornwhen;
  }
  public void setInitNumber(int birthQuantity) {
    initNum=birthQuantity;
    origInitNum=birthQuantity;
  }
  // change only working copy, not master copy
  public void setWorkingInitNumber(int i) {initNum=i;}
  public Mcr getInitArea() {if (initArea == null) return Globals.getWorldMcr(); else return this.initArea;}
  public void setInitArea(Mcr m) {this.initArea = m;}
  public int getInitNumber() {return initNum;}
  public void setPrototype(Speciesable toBeBorn) {
    progenitor=toBeBorn;
    population = 0;
  }

  public String toString() {
    return "Species[name=" + name + ",progenitor=" + progenitor +
      "\n     birthday=" + birthday + ",initNum="+origInitNum+
      ",initArea="+ ( initArea==null?"(world)":initArea.toString() );
  }
  public String shortString() {
    return "Species" + Globals.pointerString(this) + "[name=" + name + "]";
  }

  public long getBirthday() {return birthday;}
  public Speciesable getProgenitor() {return progenitor;}
  public String getName() {return name;}
  public void setName(String s){
    if (progenitor != null) {
      int colorid = progenitor.getColor();
      if (colorid >= 0) {
        Color c = QuadDisplay.getColor(colorid);
        progenitor.setColor(QuadDisplay.setColor(name, c));
      }
    }
    name=s;
    // note that all that screwing around probably doesn't change the colorid,
    // which is what QuadDisplay uses to build the legend. So, if was on the
    // legend before, it still is. Else not.
  }

  public void setInitArea(double lx, double ly, double ux, double uy) {
    initArea = new Mcr();
    initArea.setMcr(lx, ly, ux, uy);
  }

  public static long getLastBirthday() { return lastBirthday; }

  public static void stepAll() {
    for (int i=0; i<allSpecies.size(); i++)
      ((Species)allSpecies.elementAt(i)).step();
  }

  static public void stepAllStats() {
    Species s;
    for (int i=0; i<allSpecies.size(); i++) {
      s = (Species)allSpecies.elementAt(i);
      if (s.bioRate != null)
        s.bioRate.step();
      if (s.bioTrans != null)
        s.bioTrans.step();
      if (s.stages != null)
        s.stages.step();
      if (s.fragStats != null)
        s.fragStats.step();
      if (s.radii != null)
        s.radii.step();
    }
  }

  public boolean isBirthday() {
    long today = Globals.getCurrentTime();
    if (Globals.isSeasonal())
      today = today % Globals.getSeasonCycle();
    return today == birthday;
  }

  /** On birthday (and only birthday) step hangs the newborns on the world tree. */
  public void step() {
    long today = Globals.getCurrentTime();
    biomass = 0;
    if (stages != null)
      stages.zeroMembers();

    if ((progenitor != null) && isBirthday() ) {
      Speciesable lbaby;
      Mcr area = initArea;
      if (area == null) {
        area = Globals.getWorldMcr();
      }
      double rangex = area.ux - area.lx;
      double rangey = area.uy - area.ly;
      int winitNum = initNum; // let scaredygrazer change it....
      population = 0;
      for (int i=0; i<winitNum; i++) {
        lbaby = (Speciesable) progenitor.clone();
        lbaby.setX(area.lx + rangex * Globals.getRandomSource().nextDouble());
        lbaby.setY(area.ly + rangey * Globals.getRandomSource().nextDouble());
        // Ginger: DO NOT setBirthday before setX, Y. Will break Bacsim, as
        // BacSim is using this as a "hang grandma end" (001210).
        lbaby.setBirthday(today);
        lastBirthday = today;

        if (varies != null) {
          for (int j=0; j<varies.size(); j++)
            ((VaryProperty)varies.elementAt(j)).apply(lbaby);
        }

        if (lbaby instanceof Quad) {
          if (lbaby.putOnPlantRoot())
            Globals.getPlantRoot().addNewbie((Quad)lbaby);
          else
            Globals.getMobileRoot().addNewbie((Quad)lbaby);
        }

        if (lbaby instanceof HasSpeciesInit)
          ((HasSpeciesInit)lbaby).speciesInit();
          
        population += 1.0;
      }
    }
  }

  public double getPopulation() { return population; }
  public void notifyBirth() { population += 1.0; }
  public void notifyDeath() {
    population -= 1.0;
  }
  public void notifyBirthToMom(StageObservable mom){
    if (stages != null) {
      stages.tally(StageStatBuckets.KIDS, stageObserver.getStage(mom));
      // births always appear in stage 0
      stages.tally(StageStatBuckets.BIRTHS, 0);
    }
  }
  public void notifyDeath(StageObservable rip){
    if (stages != null)
      stages.tally(StageStatBuckets.DEATHS, stageObserver.getStage(rip));
  }
  public void notifyAlive(StageObservable alive){
    if (stages != null)
      stages.tally(StageStatBuckets.MEMBERS, stageObserver.getStage(alive));
    if (fragStats != null)
      fragStats.tally((QuadCircleSpecies)alive);
    if (radii != null)
      radii.tally((QuadCircleSpecies)alive);
  }
  /** At the end of each step, each member of the species notifies the Species
   * what its current biomass is. Biomass is zeroed by step.
   */
  public void notifyBiomass(double b) {
    biomass += b;
    notifyBioTrans(BiomassTransfers.BIOMASS, b);
  }
  public void notifyBioTrans(int transferType, double b) {
    if (bioRate != null)
      bioRate.addToBucket(transferType, b);
    if (bioTrans != null)
      bioTrans.addToBucket(transferType, b);
  }
  public double getBiomass() { return biomass; }

  /** Hook broker onto given plot w/ time on x-axis. */
  private void observeXTimeCommon(String plotname, StatSourceBroker broker,
                                  String tracename, Color color, boolean legendFlag)
  {
    MetaPlot mp = Globals.getStats().getMetaPlot(plotname);
    if (mp.hasTrace(tracename)) {
      return;
    }
    StatTrace st = new StatTrace(tracename, broker);
    st.setXTime(new Globals());
    st.setColor(color);
    mp.addTrace(st);
    mp.setNeedLegend(legendFlag);

    Globals.getUI().makeStatDumpers(mp);
  }

  public void observePopulation() {
    observeXTimeCommon(Globals.POP_PLOT_NAME, getPopBroker(),
                       name, QuadDisplay.getColor(progenitor.getColor()), false);
  }
  public void observeBiomass() {
    observeXTimeCommon(Globals.BIO_PLOT_NAME, getBiomassBroker(),
                       name, QuadDisplay.getColor(progenitor.getColor()), false);
  }

  /** Note the time component of fertility rates is "per step" NOT statcycle anymore. */
  public void observeFertility() {
    if (!stageObservableCheck()) return;
    observeStageCommon(Globals.STAGE_FERT_PLOT_NAME, getFertBrokers());
  }
  /** Note the time component of fertility rates is "per statcycle".
   *  Mortality is deadpercent of the members who were in the stage at
   * the beginning of the cycle, so it can't exceed 1.0.
   */
  public void observeMortality() {
    if (!stageObservableCheck()) return;
    observeStageCommon(Globals.STAGE_MORT_PLOT_NAME, getMortBrokers());
  }
  /** Stage 0 is alwasy the stage newborns go into. These stats are not
   *  divided by statcycle, so implicitly, it's per statcycle.
   */
  public void observeStages() {
    if (!stageObservableCheck()) return;
    observeStageCommon(Globals.STAGE_PLOT_NAME, getStageBrokers());
  }

  /** Observe radius mean/sigma/skew. */
  public void observeRadius() {
    if (radii == null) {
      radii = new RadiusMeanSigmaSkew(true); // do skew by default
      radiusBrokers = radii.makeBrokers();
    }
    Color[] colors = radii.getColors();
    String plotname = name + " Radius Variation";
    for (int i=0; i<radii.NUM_TRACES; i++)
      observeXTimeCommon(plotname, radiusBrokers[i],
         radii.getTraceName(i), colors[i], true);
  }

  public boolean stageObservableCheck() {
    if (progenitor instanceof StageObservable) {
      if (stageObserver == null)
        stageObserver = ((StageObservable) progenitor).getStageObserver();
      if (stages == null) {
        stages = new StageStatBuckets( stageObserver.getNumStages() );
      }
      return true;
    } else {
      Globals.getUI().log("WARNING: observeFertility/Mortality/Stages ignored--species doesn't support it.");
      return false;
    }
  }

  private void observeStageCommon(String plotsuffix, StageBucketBroker[] brokers) {
    String plotname = name + " " + plotsuffix;
    Color[] colors = stages.getColors();
    for (int i=0; i<stages.getNumStats(); i++) {
      observeXTimeCommon(plotname, brokers[i],
                         stageObserver.getStageName(i),
                         colors[i], true);
    }
  }

  public void observeBiomassTransfers() {
    observeAllTransferCommon(getBioTransBrokers(), bioTrans, Globals.BIO_TRANSFERS_PLOT_NAME);
  }
  public void observeBiomassRates() {
    observeAllTransferCommon(getBioRateBrokers(), bioRate, Globals.BIO_RATE_PLOT_NAME);
  }

  /** I don't like the net growth ratio stat. Kill it. Leave in substructure. */
  private void observeAllTransferCommon(StatSourceBroker[] brokers, BiomassTransfers biodata,
                                     String plotsuffix)
  {
    MetaPlot mp = Globals.getStats().getMetaPlot(name + " " + plotsuffix);
    String[] traceNames = biodata.getNames();
    Color[] traceColors = biodata.getColors();
    if (mp.hasTrace(traceNames[BiomassTransfers.NET_GROWTH])) {
      return;
    }

    mp.setNeedLegend(true);
    for (int i=0; i < BiomassTransfers.getNumStats(); i++) {
      if ( (i != BiomassTransfers.BIOMASS) && (i != BiomassTransfers.LAST_NET_GROWTH)
           && (i != BiomassTransfers.NET_RATIO) ) {
        StatTrace st = new StatTrace(traceNames[i], brokers[i]);
        st.setXTime(new Globals());
        st.setColor(traceColors[i]);
        mp.addTrace(st);
      }
    }
    Globals.getUI().makeStatDumpers(mp);
  }

  /** Stat vs. biomass trace. */
  private void observeSnangleCommon(int whichStat, StatSourceBroker[] brokers,
                                    BiomassTransfers biodata, String plotsuffix)
  {
    MetaPlot mp = Globals.getStats().getMetaPlot(name + " " + plotsuffix);
    String[] traceNames = biodata.getNames();
    if (mp.hasTrace(traceNames[whichStat])) {
      return;
    }

    StatTrace st = new StatTrace(traceNames[whichStat], brokers[whichStat]);
    st.addX(brokers[BiomassTransfers.BIOMASS]);
    st.setColor(QuadDisplay.getColor(progenitor.getColor()));
    mp.addTrace(st);
    Globals.getUI().makeStatDumpers(mp);
  }

  public void observeBiomassTheft() {
    observeSnangleCommon(BiomassTransfers.LOSS_STEAL, getBioTransBrokers(),
        bioTrans, Globals.BIO_THEFT_PLOT_NAME);
  }
  public void observeBiomassGrowth() {
    observeSnangleCommon(BiomassTransfers.GROWTH, getBioTransBrokers(),
        bioTrans, Globals.BIO_GROWTH_PLOT_NAME);
  }

  public PopBroker getPopBroker(){
    if (popBroker == null)
      popBroker = new PopBroker(this);
    return popBroker;
  }
  public BiomassBroker getBiomassBroker(){
    if (biomassBroker == null)
      biomassBroker = new BiomassBroker(this, progenitor.getBiomassResourceId());
    return biomassBroker;
  }

  public BiomassBucketBroker[] getBioTransBrokers(){
    if (bioTrans == null)
      bioTrans = new BiomassTransfers(false);
    if (bioTransBrokers == null)
      bioTransBrokers = bioTrans.makeBrokers(progenitor.getBiomassResourceId());
    return bioTransBrokers;
  }

  public BiomassBucketBroker[] getBioRateBrokers(){
    if (bioRate == null)
      bioRate = new BiomassTransfers(true);
    if (bioRateBrokers == null)
      bioRateBrokers = bioRate.makeBrokers(progenitor.getBiomassResourceId());
    return bioRateBrokers;
  }


  public StageBucketBroker[] getFertBrokers() {
    if (!stageObservableCheck())
      return null;
    if (fertBrokers == null)
      fertBrokers = stages.makeFertBrokers();
    return fertBrokers;
  }
  public StageBucketBroker[] getMortBrokers() {
    if (!stageObservableCheck())
      return null;
    if (mortBrokers == null)
      mortBrokers = stages.makeMortBrokers();
    return mortBrokers;
  }
  public StageBucketBroker[] getStageBrokers() {
    if (!stageObservableCheck())
      return null;
    if (stageBrokers == null)
      stageBrokers = stages.makePopPercentBrokers();
    return stageBrokers;
  }

}
