//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990903, vfb, original Java
 */
/** Take a picture of the display. Whatever "picture" means. First up is
 *  POVray input (.pov file output). POV-Ray is a "Persistence of Vision
 *  ray tracer". Using it to make bubble picture output.
 */


package gecko.camera;
import gecko.*;
import gecko.Steppable;
import gecko.zoo.*;
import stats.*;
import java.io.*;
import java.util.Vector;

public class POVCamera implements Steppable {
  private String fileStem;
  private String fileSuffix;
  private Mcr field;
  private String globalsInclude = "grassworlds.inc"; // default is only for example

  /** The globalsFile is the path of the macros #include'd for this protocol.
   *  That #include file in the POVray file should set up the background
   *  and camera, and macros for each species, by species name, with the
   *  macros taking 4 arguments (centerx,y,z, radius). The desirable colors
   *  for a 3-D bubble picture are highly unlikely to match the colors desired
   *  on a nearly black background for circles. One exception: any descendent
   *  of Carnie is written out as HSpecies or SSpecies, depending on whether it
   *  is awake, because I haven't managed yet to get conditionals working in
   *  a macro. *Sigh*. The documentation says it's ok....
   */
  public POVCamera(String globalsFile) {
    fileStem = ( (BatchUI) Globals.getUI()).getFileStem();
    fileSuffix = getSuffix();
    globalsInclude = globalsFile;
    field = Globals.getWorldMcr();
  }

  public void setField(double xmin, double ymin, double xmax, double ymax) {
    Mcr anMcr = new Mcr();
    anMcr.setMcr(xmin, ymin, xmax, ymax);
    setField(anMcr);
  }
  public void setField(Mcr mcr) {
    field = mcr;
  }

  public String getSuffix() { return ".pov";}

  public void step() {
    long time = Globals.getCurrentTime();
    String filename = fileStem + "_t" + time + fileSuffix;
    PrintWriter newOutput = Globals.getUI().openFile(filename);
    writeAll(newOutput);
    newOutput.flush();
    newOutput.close();
  }

  public void writeAll(PrintWriter out) {
    writeGlobals(out);
    writeTree(out, Globals.getWorldRoot());
    writeEnd(out);
  }
  /** nothing at present--a place to write that-which-we-don't-know-until-looking */
  public void writeEnd(PrintWriter out) {  }

  /** The data recursion..*/
  private void writeTree(PrintWriter out, Quad root) {
    // These trees can be vast--don't do both directions recursively....
    // And optimize every bit possible.
    Quad thislevel = root;
    QuadCircleSpecies c;

    if (root == null)
      return;

    do {
      if (thislevel.child != null)
        writeTree(out, thislevel.child);
      if (thislevel instanceof QuadCircleSpecies) {
        writeCritter( out, (QuadCircleSpecies) thislevel );
      }
    } while ( (thislevel = thislevel.sibling) != null );
  }


  /** For POVray output, can either be vastly verbose, or set up
   *  macros for the critters we want to draw. writeGlobals and
   *  writeCritter are thus inextricably linked, since writeCritter
   *  is calling macros built by writeGlobals.
   *
   *  This version creates sky, Plant, Plant2, Plant3, Gopper, HSpider,
   *  and SSpider for use by writeCritter. Also generates the sky and
   *  such.
   *
   *  This routine also invokes macros Time(int) and WorldSize(minx,miny,maxx,maxy)
   *  in case the globals.inc file wants to adjust the picture accordingly.
   */
  public void writeGlobals(PrintWriter out) {
    out.println("// time " + Globals.getCurrentTime());
    out.println("#include \"" + globalsInclude + "\"");

    out.println("TimePicture (" +  Globals.getCurrentTime() + ")");
    Mcr landsize = field;
    out.println("WorldSize(" + landsize.lx + "," + landsize.ly +","
				     + landsize.ux + "," + landsize.uy + ")");
  }

  public void writeCritter(PrintWriter out, QuadCircleSpecies critter) {
    String povobject = critter.getSpecies().getName();
    povobject.replace(' ','_');
    double height;

    // skip this critter if it doesn't cross the field
    if ( !field.crossMcr(critter) )
      return;

    if (critter instanceof Carnie) {
      if ( ((Carnie)critter).isSleeping() )
        povobject = "S" + povobject;
      else
        povobject = "H" + povobject;
    }

    // hard-code some heights, since z not meaningful in grassworlds
    if (critter instanceof Carnie) {
      height = critter.getRadius();
    } else {
      height = critter.getZ();
    }
    
    out.println(povobject + "(" +
                critter.getX() + "," +
                critter.getY() + "," +
                height + "," +
                critter.getRadius() + ")"
    );
  }

}