//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator

/** Stage statistics involve one sorting-bucket set for each of the
 *  lifecycle stages of a beastie: number alive, number died, number
 *  born from a member, number of births into stage. The stats derived
 *  from this are %population by stage, fertility rate per capita per statcycle
 *  per stage, same for mortality.
 *
 *  These stats are popular in ecology, often called l(x)m(x) tables or something.
 *  At any event, it's a stat that may be available from field data for a
 *  system, thus interesting for checking simulation lifecycle verisimilitude.
 *
 *  SwarmGecko had a bunch of bucket stats (age, stage, specific, etc.) of which
 *  never more than one was used at a time. For this rewrite, just allow one
 *  "stage" stat, and may the user/species author implement it as needed.
 */

package gecko.statglue;

import gecko.Globals;
import stats.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.awt.Color;

public class StageStatBuckets extends StatSource {
  // these are observation counters
  static public final int MEMBERS          = 0; // at end of statcycle?
  static public final int KIDS             = 1; // to members of the stage
  static public final int BIRTHS           = 2; // only in stage 0
  static public final int DEATHS           = 3; // of members of the stage

  // these are derived, not actually used in inBuckets
  static public final int FERT             = 4;
  static public final int MORT             = 5;
  static public final int POP_PERCENT      = 6;
  static final private int numStats = 7;
  private StatBuckets[] stages;
  private int numStages;
  static private Color[] stageColors = null;

  public StageStatBuckets(int numStages) {
    int i;
    this.numStages = numStages;
    stages = new StatBuckets[numStages];
    for (i=0; i<numStages; i++)
      stages[i] = new StatBuckets(numStats);

    // let's go ahead and provide color service for these stages
    if (stageColors != null)
      return;
    Palette p = new Palette(numStages, Palette.PASTEL);
    stageColors = new Color[numStages];
    for (i=0; i<numStages; i++)
      stageColors[i] = p.getColor(i);
  }
  public int getNumStats() {return numStages;}
  static public Color[] getColors() { return stageColors; }
  public double getStat(int stage, int stat) { return stages[stage].getBucket(stat);}

  public void tally(int whichStat, int whichStage) {
    stages[whichStage].incrementBucket(whichStat);
  }
  /** Not really any convenient place to notifyAlive except critter step. So
   *  every step (not just stat steps) need to zero out members. */
  public void zeroMembers() {
    for (int i=0; i<numStages; i++)
      stages[i].overrideInBucket(MEMBERS, 0.0);
  }
  /** Promote inBucket to outBucket, then calculate the derived stats.
   *  vfb, changed interpretation, 991027, so that statcycle doesn't affect answers, and
   *  answers correctly reflect how agents begin each step. Changed here, AND
   *  everywhere notifyAlive(StageObservable) was called (begin step on agent).
   */
  public void step() {
    int a;
    double members, total, membersCycle;
    total = members = 0;
    StatBuckets stage;

    // simple sums -- total population includes births
    for (a=0; a < numStages; a++) {
      stages[a].step();
      total += stages[a].getBucket(MEMBERS);
    }

    // probabilities
    for (a=0; a < numStages; a++) {
      stage = stages[a];
      // members is those who begin statcycle in this stage
      members =   stage.getBucket(MEMBERS);
      membersCycle = members * Globals.getStatCycle();
      if (membersCycle > 0) {
        stage.setBucket(FERT, stage.getBucket(KIDS) / membersCycle);
        stage.setBucket(MORT, stage.getBucket(DEATHS) / membersCycle);
      } else { // undefined, should be 0
        stage.setBucket(FERT, 0);
        stage.setBucket(MORT, 0);
      }
      if (total > 0)
        stage.setBucket(POP_PERCENT, members / total);
      else
        stage.setBucket(POP_PERCENT, 0);
    }
  }

  /** There's actually a matrix of brokers one might want. So,
   *  split up by FERT, MORT, and POP_PERCENT arrays of brokers[stage].
   */
  private StageBucketBroker[] makeBrokers(int whichStat) {
    StageBucketBroker[] brokers = new StageBucketBroker[numStages];
    for (int i=0; i < numStages; i++) {
      brokers[i] = new StageBucketBroker(this, i, whichStat);
    }
    return brokers;
  }
  public StageBucketBroker[] makeFertBrokers(){return makeBrokers(FERT);}
  public StageBucketBroker[] makeMortBrokers(){return makeBrokers(MORT);}
  public StageBucketBroker[] makePopPercentBrokers() {return makeBrokers(POP_PERCENT);}
}
