//Title:        Gecko Woods
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 991108, vfb, original Java
 * This is a helper class for Tree. Simply subtracting overlap*lightCoeff,
 * in the case of light, is significantly overstated, because light filtering is
 * multiplicative, not additive. So let's bite the bullet and make sure we get
 * the same results if we do it the tedious way.
 *
 * 991109, vfb, added flag (in Woods) to say whether to use cheaper diffusive
 * lightbox. In that twist, the debit is multiplied into the entire area, so
 * no subboxes are kept. The way it's implemented, it's quite possible to flip
 * the switch back and forth (some operations boxy, some others diffusive) on
 * the same lightbox in the same calculation.
 */


package gecko.woods;
import gecko.*;
import java.util.Vector;

public class Lightbox {
  // NOTE:
  private BoxDensity firstBox;
  private Vector boxes; // of BoxDensity
  private int chunk = 25;

  public Lightbox(Mcr begin, double initLightDensity) {
    firstBox = new BoxDensity(begin, initLightDensity);
    Mcr firstChopper = new Mcr();
    firstChopper.setFrom(begin);
    boxes = new Vector(chunk, chunk);
    boxes.addElement(new BoxDensity(firstChopper, initLightDensity));
  }

  /* This version only makes sense on diffusive lightboxes. */
  public void factorIn(double area, double lightDensity) {
    BoxDensity boxd = ((BoxDensity) boxes.elementAt(0));
    double bigArea = boxd.mcr.getArea();
    double littleArea = area;
    double areaDens = (bigArea - littleArea) * boxd.density
                      + (littleArea * lightDensity * boxd.density);
    boxd.density = areaDens / bigArea;
  }

  public void factorIn(Mcr cross, double lightDensity) {
    Mcr crossm = new Mcr();
    // adding new boxes to end, past where we begin....
    for (int i=boxes.size() - 1; i>=0; i--) {
      BoxDensity boxd = ((BoxDensity) boxes.elementAt(i));
      Mcr fragment = boxd.mcr;
      int crossmtype = cross.intersectMcr(fragment, crossm);
      switch (crossmtype) {
        case Mcr.MCR_DISJOINT: break;
        case Mcr.MCR_CONTAINS: boxd.density *= lightDensity;  break;
        default:
          // note that in diffusive model, there will only ever be
          // one box. Simplifies code to keep the structure, tho.
          if (Woods.isDiffusiveLightBox()) {
            double bigArea = boxd.mcr.getArea();
            double littleArea = crossm.getArea();
            double areaDens = (bigArea - littleArea) * boxd.density
                              + (littleArea * lightDensity * boxd.density);
            boxd.density = areaDens / bigArea;
          } else {
            breakBox(boxes, boxd, crossm, lightDensity);
          }
      }
    }
  }

  // The result of all this nonsense is the sum of the products of
  // all the little boxes.
  public double getLightDensity() {
    double answer = firstBox.mcr.getArea();
    if (answer == 0)
      return 0; // no propagating NaN's, please....

    double lightArea = 0;
    BoxDensity boxd;
    for (int i=0; i<boxes.size(); i++) {
      boxd = ((BoxDensity)boxes.elementAt(i));
      lightArea += boxd.mcr.getArea() * boxd.density;
    }
    answer = lightArea / answer;
/* v. useful debug */
/*
if (Globals.getRandomDouble() < .01) {
System.err.println("getLightDensity=" + answer + " over " + firstBox.mcr);
for (int i=0; i<boxes.size(); i++) {
  boxd = ((BoxDensity)boxes.elementAt(i));
  System.err.println("Box " + i + " : " + boxd.density + " over " + boxd.mcr);
}
}
*/

    if (Double.isNaN(answer))
      answer = 0;
    return answer;
  }

  public void breakBox(Vector boxes, BoxDensity toBreak, Mcr cross, double lightDensity) {
    // toBreak is going to be replaced in situ with the cross Mcr, because
    // that's the only one that's supposedly never null. In the general case,
    // with cross completely within toBreak box, the result of breaking is
    // 5 BoxDensity boxes. But use fewer whenever possible....which is usually.
    Mcr newToBreak = new Mcr();
    newToBreak.setFrom(cross);
    // but store it last.

    // below
    Mcr abox = new Mcr();
    abox.setFrom(toBreak.mcr);
    abox.uy = cross.ly;
    if (abox.getArea() > 0)
      boxes.addElement(new BoxDensity(abox, toBreak.density));

    // above
    abox = new Mcr();
    abox.setFrom(toBreak.mcr);
    abox.ly = cross.uy;
    if (abox.getArea() > 0)
      boxes.addElement(new BoxDensity(abox, toBreak.density));

    // left
    abox = new Mcr();
    abox.lx = toBreak.mcr.lx;
    abox.ux = cross.lx;
    abox.ly = cross.ly;
    abox.uy = cross.uy;
    if (abox.getArea() > 0)
      boxes.addElement(new BoxDensity(abox, toBreak.density));

    // right
    abox = new Mcr();
    abox.lx = cross.ux;
    abox.ux = toBreak.mcr.ux;
    abox.ly = cross.ly;
    abox.uy = cross.uy;
    if (abox.getArea() > 0)
      boxes.addElement(new BoxDensity(abox, toBreak.density));

    toBreak.mcr = newToBreak;
    toBreak.density = lightDensity * toBreak.density;
  }

  class BoxDensity {
    Mcr mcr;
    double density;

    public BoxDensity(Mcr m, double dense) {
      mcr = m;
      density = dense;
    }
  }
}

