//Title:        Gecko Woods
//Version:
//Copyright:    Copyright (c) 2000, All rights reserved
//Author:       Ginger Booth, ginger@gingerbooth.com, gingerbooth@cyber-wizard.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 000831: vfb: original didn't work worth a damn. Let's try combining Tree2's
 * partition scheme (but use the progressive partitions!) with Tree3's crown
 * footprint competition. Retire crownToStem altogether--have init and maxCrown
 * sizes. And leafturnover.
 *
 * 000824, original: pull invariant species parameters out of Tree3, to cut
 * memory needs. All fields are communicated via Tree3, not here. Package-scope
 * publicity. Delegate record of Tree3.
 */
package gecko.woods;
import gecko.*;
import java.awt.Color;

public class Tree3Params {
  // defining parameters for the species
  // see p. 11, Pacala et al 1996 Ecological Monographs.
  double maxHeight;         // H1--meters
  double initHeight;   // H2--meters/cm
  double height50;    // height at 50 cm dbh
  double crownDepthToHeight;// C1--m/m
  double lightCoeff;        // exp(-Ei), as given in table p. 11
  double recruitMeanD;      // mean distance parent->recruit (instead of R1)
  double bigFertility;      // R2

  // these params are Tree3's invention, not from Sortie (tho values may be)
  double minRadialIncrement; // cm, min growth in dbh radius--like ring width, per year
  int    yearsForMRI = 5;   // not on ui--may never be overridden
  double initCrownRadius;   // m to begin
  double crownToStem;       // describes crown growth

  // partition scheme
  double woodDensity;       // density of wood (convert productivity tonnes -> volume)
  double stemPartBegin;     // stem partition of productivity, at init size
  double stemPartEnd;       // stem partition of productivity, at max height

  double mortRandom;        // no name given - applies to the SPECIES
                            // Reaper does global mortality games

  // these are commonly allowed to stay at these given values--not on params menu
  double recruitDBH = 1.0;    // cm
  double dbh50 = 50.0;    // cm
  int recruitAge = 5;         // age at which newborn starts stepping

  // derived values must be set after all the defining params are set
  double stemSlope;

  public double zFormula(double dbh) {
    // 000908, vfb: think a parabolic formula will fix weird ba-peaking thing.
    double zcoeff = (height50 - initHeight) / Math.sqrt(dbh50 - recruitDBH);
    return initHeight + zcoeff * Math.sqrt(dbh - recruitDBH);

    // older Sortie formula
    // return maxHeight * (1.0 - Math.exp(- dbh * heightInitSlope/maxHeight));
    // Canham's formula--don't use, as tree misshapen for our purposes
    //    return heightOffset + (maxHeight - heightOffset) *
    //      (1.0 - Math.exp(- dbh * heightInitSlope/maxHeight));
  }

  /** Just a utility routine. */
  static public double coneVolume(double r, double h) {
    return (Math.PI / 3.0) * r * r * h;
  }
  /** convert back, return -radius for -volume */
  static public double coneRadius(double vol, double h) {
    double inside = (vol * 3.0) / (h * Math.PI);
    double ans = Math.sqrt( Math.abs( (vol * 3.0) / (h * Math.PI) ) );
    if (inside < 0)
      ans = - ans;
    return ans;
  }

  /** set up derived values after defining params set */
  public void setDerived() {
    initHeight = zFormula(recruitDBH);

    double rise = stemPartEnd - stemPartBegin;
    double run = maxHeight - initHeight;
    stemSlope = rise/run;
  }

  /** freshWood is the living part of the stem. Since how much of the crown
   *  is included in "stem" matters in several places, localize the scheme
   *  du jour. Density 1.0 this-wood-density (whatever that is).
   * NOTE: deltaBA is in cm since stemR is kept in cm. Returns m^3.
   * Ignore change in height, assume nominal at this point in growth.
   */
  static public double getFreshWood(double height, double crownDepth, double deltaBA) {
    // hollow cyclinder, delta-basal-area * height. The height used is to halfway
    // up the crown.
    double freshWood = (height - crownDepth/2.0) * deltaBA;
    freshWood /= 10000; // because ba is in cm and height is in m.
    return freshWood;
  }
  /** the un-zFormula */
  /*
  public double getDbhForHeight(double height) {
    return  - Math.log(1.0 - height/maxHeight) / (heightInitSlope/maxHeight);
  }
  */

  /** fullBudget is in m^3 of this-density */
  public double getGrowthBudget(Tree3 tree, double effArea) {
    double kgBudget = effArea * tree.getNutrientFactor();
    // nutrientFactor is in tonne/ha, effArea is in m^2,
    // there are 10,000 m^2/ha and 1,000 kg/tonne, so ...  1000/10000 = 1/10
    kgBudget /= 10;
// System.err.println("effArea=" + effArea + ", nutFac=" + tree.getNutrientFactor()
// + ", kgBudget=" + kgBudget);
    return kgBudget / woodDensity;
  }
  /** stemBudget is in m^3 of this-density. */
  public double getStemPart(Tree3 tree, double fullBudget) {
    // where are we on the stem partition schedule?
    // leave off the stunting thing for now--ideally, should shift up-schedule for height
    double xCoord = tree.z - initHeight;
    double partition = stemSlope * xCoord + stemPartBegin;
    return fullBudget * partition;
  }

}
