//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990329, vfb, original Java
 * Translated from the Swarm Objective-C Gecko source, version 0.51
 * With structural changes.
 */
/** One basic living thing ala Gecko's worldview (metabolism, lifecycle, etc.)
 *  Note the equivalent of SwarmGecko's old Critter is now SizedPlant.
 */


package gecko.zoo;

import java.awt.Color;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import gecko.statglue.*;
import gecko.*;

public abstract class Critter extends QuadCircleSpecies implements StageObservable, GetDirectionable {
  // defining settings
  private Metabolism metabolism; 	  // stores, gullet, conversions, taxes, etc.
  private double scaleradius;	  // all radii = scale * (radius from volume)
  private double scalev;	      // volume = scalev * (sphere volume)
  private int maxAge;           // maximum lifespan, 0 if don't die of old age.

  /** This property list is used to generate paramLists for popup editing.
   *  Note initArea is left off for now, as don't have the routines to set
   *  it nicely from a popup.
   */
  protected static String[] props = { "radius", "maxAge", "scaleRadius", "scaleVolume",
    "conversion", "taxExponent", "taxRate" };

  // working settings
  // color, xy, and radius are public only so that QuadDisplay can go faster.
  private double lastr;       // last size...
  protected Direction direction = new Direction();  // scratch space, where we're headed
  private boolean plantRootFlag = true; // passive until told otherwise

  // scratchpad moved to PassiveCritter

  // life history stats...dunno if I want all these in the long run....
  protected long deathday;	  // day to expire, 0 if no maxAge.
  protected long firstbornon; // getCurrentTime() at first kid
  protected int atecount;	  // meals I've had
  protected int kidcount;		  // kids I've had

  public long getFirstBornOn() {return firstbornon;} // for stage observer's use

  public Critter() {
    super();
    init();
    deathday = -1;
  }
  public void setMetabolism(Metabolism m) {metabolism = m;};
  public Metabolism getMetabolism() {return metabolism;}
  public int getBiomassResourceId() {return metabolism.getIam();}
  /** This routine is actually static, but i/f usage won't allow it to be...*/
  public String[] getPropertyNames() { return props;}

  public void setBirthday(int i) { birthday = i;}
  /** Don't set a maxAge and also kill them onSeason--let onSeason kill them
   *  in that event. For the Basic3 scenarios, maxAge = 0 (never die of old age.)
   */
  public void setMaxAge(int i) { maxAge = i;}
  public int getMaxAge() { return maxAge;}
  public boolean hasMaxAge() { return (maxAge > 0);}

  /** Critters default to plant metabolism, all scales = 1. */
  public void init() {
    metabolism = new Metabolism();
    metabolism.setIeat(GrassResources.RESOURCE_SUNWATER);
    metabolism.setIam(GrassResources.RESOURCE_CARBOS);
    setType(QUAD_CRITTER);
    scaleradius = 1;
    scalev = 1;
    maxAge = 0;
  }

  public void setRadius(double r) {
    radius = r;
    matchVolumeToRadius();
  }

  /** For being born only. Note this only sets heritable qualities and
   * the location. Quadness unintialized.
   */
  public void setFrom(Critter mom) {
    super.setFrom(mom);
    metabolism.setFrom(mom.metabolism);
    maxAge = mom.maxAge;
    setType(QUAD_CRITTER);
    scaleradius = mom.scaleradius;
    scalev = mom.scalev;
    matchVolumeToRadius();
    if (maxAge > 0)
      deathday = getBirthday() + maxAge;
    else
      deathday = -1;
    stageObserver = mom.stageObserver;
  }

  /** Note, this generally isn't called. QuadDisplay knows about Critters and
   *  draws them directly for performance reasons.
   */
  public void paint(QuadDisplay qd) {
    qd.drawCircle(color, x, y, radius, false);
  }

  abstract public boolean wantsToReproduce();
  abstract public ResourceList getKidShare();
  abstract public double getKidDistance();
  abstract public Move getDesiredMove();
  abstract public boolean isStarving();

  /** FeedQuad needs access to protected member.*/
  public void addDirection(double dx, double dy, double weight) {
    direction.addDirection(dx, dy, weight);
  }
  /** Return Direction vector for others to play with. */
  public Direction getDirection() { return direction; }

  /** False if didn't. */
  public boolean reproduce() {
    if (!wantsToReproduce())
      return false;

    Critter baby;
    double distance;
    double angle;
    ResourceList kidgets = getKidShare();

    if (kidcount == 0)
      firstbornon = Globals.getCurrentTime();
    kidcount++;

    angle = 2.0 * Math.PI * Globals.getRandomSource().nextDouble();
    // make the kid
    Speciesable lbaby = makeKid();
    baby = (Critter) lbaby;
    distance    = getKidDistance();
    baby.x      = x + ( distance * Math.cos(angle) );
    baby.y      = y + ( distance * Math.sin(angle) );
    baby.direction.setDirection(angle);

    // transfer resources
    metabolism.debit(kidgets);
    resetRadius();
    baby.metabolism.setStores(kidgets);
    baby.resetRadius();

    // hang on quad tree
    baby.setMcr();
    Quad myroot = getRoot();
    myroot.addNewbie(baby);
    getSpecies().notifyBirth();

    return true;
  }

  abstract public Object clone();

  /** by default, Critters die at the beginning of step on deathday, if any
   *  deathday specified. Returns true if died.
   */
  public boolean deathdayCheck() {
    if (hasMaxAge()) {
      if (deathday == Globals.getCurrentTime()) {
        die();
        return true;
      }
    }
    return false;
  }

  /** by default, Critters do nothing on hunt. */
  public void hunt() {  }

  public void move() {
    Move move = getDesiredMove();
    if ((move.x == 0.0) && (move.y == 0))
      return;

    x += move.x;
    y += move.y;
    direction.zeroDirection();	// done with this intention.
    setMcr();
  }

  /** Default Critter does nothing onSeason. */
  public boolean onSeason() {return false;}
  /** Calls onSeason on seasonCycle. Which then does nothing. */
  public boolean seasonCheck() {
    if (Globals.isSeasonNow())
      return onSeason();
    return false;
  }

  /** One time step's worth of lifecycle...whole life happens step by step.
   *  Order is:
   *    0. check deathday and season (added 990809)
   *    1. metabolize
   *    2. paytaxes (die)
   *    3. reproduce
   *       3b. if DIDN'T reproduce, then hunt
   *    4. move (yes, plants DO move)
   *
   * vfb, 990809: adding deathday and onSeason, realized there's no good reason
   * for Simulator to be involved in that stuff. Can implement here.
   */
  public void step() {
    // for some reason, did not step on birthdays past first season...dunno why,
    // but it makes subsequent seasons unmatch first in that seedings remove overlap
    // on first season and not subsequent ones.... Because plant got fed on birthday,
    // but didn't get a chance to die, so got fed twice before paying any taxes.

    getSpecies().notifyAlive(this); // for stages
    if (deathdayCheck())
      return;
    if (seasonCheck())
      return;
    metabolize();
    if (!paytax()) {
      return;
    }
    if (!reproduce()) {
      hunt();
    }
    move();
    getSpecies().notifyBiomass(metabolism.getStores(metabolism.getIam()));
  }

  /** Last radius is a saved off copy. */
  protected double getLastRadius() {return lastr;}
  public void matchVolumeToRadius() {
    double volume = Globals.radius2volume(radius, scaleradius, scalev);
    int Iam = metabolism.getIam();
    metabolism.setStores(Iam, volume);
  }

  public double getScaleVolume() {return scalev;}
  public void setScaleVolume(double s) {scalev = s;}
  public long getAge() { return Globals.getCurrentTime() - birthday; }
  public long getDeathday() { return deathday;}

  private static StageObserver stageObserver = new DefaultStageObserver();
  public StageObserver getStageObserver(){
    return stageObserver;
  }

  public boolean putOnPlantRoot() {return plantRootFlag;}
  public void setPutOnPlantRoot(boolean flag) {plantRootFlag = flag;}

  /** Note: deathday is a working setting, not a species parameter. If you wish to
   *  say the species has a lifespan of 200, need to setMaxAge(200) instead.
   */
  protected void setDeathday(int dieOn) { deathday = dieOn;}

   /** Find QUAD_SESSILE or QUAD_MOBILE this guy's hanging below. */
  public Quad getRoot() {
    Quad myRoot = getParent();
    while ( (myRoot != null) &&
            (myRoot.getType() != QUAD_SESSILE) &&
            (myRoot.getType() != QUAD_MOBILE)
          )
      myRoot = myRoot.getParent();
    return myRoot;
  }

  public void setXYR(double nx, double ny, double nr) {
    x = nx; y = ny;
    setRadius(nr);
    setMcr();
  }
  public void setXY(double nx, double ny) {
    x = nx; y = ny;
    setMcr();
  }
  public void setXYZ(double nx, double ny, double nz) {
    x = nx; y = ny; z = nz;
    setMcr();
  }
  public void setXYZR(double nx, double ny, double nz, double nr) {
    x = nx; y = ny; z = nz; radius = nr;
    setMcr();
  }

  public void setConversion(double rate) {

    metabolism.setConversion(metabolism.getIeat(), metabolism.getIam(), rate);

  }
  public double getConversion() {
    return metabolism.getConversion(metabolism.getIeat(), metabolism.getIam());
  }

  public void setScaleRadius(double d) {scaleradius = d;}
  public double getScaleRadius() {return scaleradius;}

  public void setTaxRate(double rate){
    metabolism.setTax(metabolism.getIam(), rate, metabolism.getTaxExponent(metabolism.getIam()));
  }
  public double getTaxRate() {
    return metabolism.getTaxRate(metabolism.getIam());
  }

  public void setTaxExponent(double exp){
    metabolism.setTax(metabolism.getIam(), metabolism.getTaxRate(metabolism.getIam()), exp);
  }
  public double getTaxExponent() {
    return metabolism.getTaxExponent(metabolism.getIam());
  }

  /** if we eat the same thing, we compete. For a default.... */
  public boolean competesWith(Critter other) {
    int sheEats = other.metabolism.getIeat();
    if ( sheEats > GrassResources.RESOURCE_NULL) {
      if ( metabolism.doIeat(sheEats) ) {
        return true;
      } else {
        return false;
      }
    } else {
      for (int i=1; i < Globals.getResources().getNumResources(); i++) {
        if ( metabolism.doIeat(i) && other.metabolism.doIeat(i) )
          return true;
      }
    }
    return false;
  }

  /** if I eat what she converts food into, I want to eat her. */
  public boolean eats(Critter other){
    int sheis = other.metabolism.getIam();
    if (sheis > GrassResources.RESOURCE_NULL) {
      if (metabolism.doIeat(sheis))
        return true;
      else
        return false;
    } else {
      for (int i=1; i < Globals.getResources().getNumResources(); i++) {
        if ( metabolism.doIeat(i) && other.metabolism.amI(i) )
          return true;
      }
    }
    return false;
  }

  public boolean eats(int resource){ return metabolism.doIeat(resource); }
  public boolean attacks(Critter other){
    return metabolism.doIeat(other.metabolism.getIam());
  }

  public boolean eats(ResourceList list) {
    int meat = metabolism.getIeat(); // me eat
    if (meat > GrassResources.RESOURCE_NULL) {
      if (list.list[meat] > 0.0)
        return true;
      else
        return false;
    } else {
      for (int i=1; i < Globals.getResources().getNumResources(); i++) {
        if ( metabolism.doIeat(i) && (list.list[i] > 0.0))
          return true;
      }
    }
    return false;
  }

  public boolean makes(int resource){ return metabolism.amI(resource); }

  /** debit this Critter this amount of this resource--caller is stealing.*/
  public void stealResource(int resourceType, double amount)
  {
    double stores = metabolism.getStores(resourceType);
    stores -= amount;
    metabolism.setStores(resourceType, stores);
    getSpecies().notifyBioTrans(BiomassTransfers.LOSS_STEAL, amount);
  }

  public double stealAllResource(int resourceType) {
    double stores = metabolism.getStores(resourceType);
    metabolism.setStores(resourceType, 0);
    getSpecies().notifyBioTrans(BiomassTransfers.LOSS_STEAL, stores);
    return stores;
  }

  public double getVolume() {return metabolism.getStores(metabolism.getIam());}

  public void feed(int resource, double amount) {
    metabolism.feed(resource, amount);
    if (amount > 0.0)
      atecount++;
  }

  public void remove(int resource, double amount){
    metabolism.setStores(resource,
      metabolism.getStores(resource) - amount);
  }

  /** returns how much was removed */
  public double removeAll(int resource) {
    double stores = metabolism.getStores(resource);
    metabolism.setStores(resource, 0);
    return stores;
  }

  public void metabolize(){
    lastr = radius; // save off what size we were...will reset at grow step
    metabolism.metabolize();
  }

  /** set size Globalsd on current stores */
  public void resetRadius() {
    int Iam, i;
    Iam = metabolism.getIam();
    if (Iam > 0) {
      radius = Globals.volume2radius(metabolism.getStores(Iam), scaleradius, scalev);
    } else {
      // well, we have no -good- definition.  So let's add 'em up....
      double vol = 0;
      for (i = 1; i < Globals.getResources().getNumResources(); i++)
        vol += metabolism.getStores(i);
      radius = Globals.volume2radius(vol, scaleradius, scalev);
    }
    setMcr();
  }

  /** Death is a side effect of not being able to pay taxes.... If routine
   *  returns false, Critter died.
   */
  public boolean paytax() {
    double netgrowth;
    int mebody = metabolism.getIam();
    ResourceList taxes = metabolism.payTaxes();
    Globals.getTaxCollector().payTax(x, y, taxes);

    resetRadius();
    netgrowth = metabolism.getStores(mebody);	// current
    netgrowth -= Globals.radius2volume(lastr, scaleradius, scalev);	// last
    getSpecies().notifyBioTrans(BiomassTransfers.GROWTH, netgrowth);

    if (isStarving()) {
      die();
      return false;
    }
    return true;
  }

  /** Note that deaths from starvation or senescence both get tallied as
   *  LOSS_STARVE in biomassTransfers statistics.
   */
  public void die() {
    double biomass = metabolism.getStores(metabolism.getIam());
    Globals.getTaxCollector().payTax(x, y, metabolism.getStores());
    remove();
    Species s = getSpecies();
    s.notifyDeath();  // for population
    s.notifyDeath(this); // for stage
    s.notifyBioTrans(BiomassTransfers.LOSS_STARVE, biomass);
  }


  public String toString() {
    return "Critter[" + "x=" + TruncVal.truncVal(x)

      + ",y=" +TruncVal.truncVal(y)

      + ",radius=" + TruncVal.truncVal(radius)

      + ",metabolism=" + metabolism

      + ",over" + ((Quad)super).toString() + "]";

  }

}

