//Title:        Gecko (class created for Woods)
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 991220, vfb, original Java
 * This weird little class tries to find an array of "random boxes" covering
 * a given area, a given target %area to cover, and min and max edges of boxes.
 * Given the same input parameters, it will always generate the same pattern of
 * boxes. However, varying an input parameter slightly (+.01) will give
 * a new random pattern. Counterintuitively perhaps, this process is perfectly
 * happy with %area less than or greater than 1.0 (>1 means they overlap), but
 * is highly unlikely to succeed if asked to give a coverage of 1.0. Another
 * sore spot is using a max edge that's "big" relative to the dimensions of area
 * to cover (like, over half). The proud result of all this is a list of Mcr's
 * giving the randomly placed boxes.
 */

package gecko.zoo;
import gecko.*;
import java.util.*;
import courseware.util.MemCleaner;
import java.awt.*;

public class RandomMcrs {
  static private Mcr overArea;
  static private double cover;
  static private double minEdge;
  static private double maxEdge;
  static private Random randomSource;
  static private Vector boxList;
  // for debug/watching--only works in application mode
  // set naptime in placeBoxes to adjust speed
  static public boolean watch = false;
  static private Frame watchFrame = null;
  static private QuadDisplay quadd;

  /** return the random number stream used to create the boxes. Used
   *  by Woods FragSite in preference to subclassing this to assign
   *  the quality of the site from the same reproduceable series.
   */
  public static Random getLastRandomSource() {return randomSource;}

  /** returns Vector of Mcr */
  public static Vector makeRandomMcrs(Mcr area, double percent,
                                      double minedge, double maxedge)
  {
    overArea = area;
    cover    = percent;
    minEdge  = minedge;
    maxEdge  = maxedge;

    // 0. initialize new random stream for repeatability
    double seed = overArea.getLenX() * overArea.getLenY() * cover
      * minEdge * maxEdge;
    if (seed < 10000)
      seed *= 10000;
    randomSource = new Random(Math.round(seed));

    // 1. make box collection
    makeBoxes();

    // 2. sort the boxes, need to place biggest dimensions first
    sortBoxes();

    // 3. Place/de-overlap boxes.
    placeBoxes();

    // 4. Return best boxes.
    return boxList;
  }

  static private void makeBoxes() {
    double cumArea = 0;
    double targetArea = cover * overArea.getLenX() * overArea.getLenY();
    double minBoxArea = minEdge * minEdge;
    double edgeRange = maxEdge - minEdge;
    boxList = new Vector(10,10);
    Mcr abox;
    double aboxArea;
    while (cumArea < (targetArea - minBoxArea)) {
      abox = new Mcr();
      abox.lx = abox.ly = 0; // default Mcr is inverted, not zeroed.
      abox.ux = (randomSource.nextDouble() * edgeRange) + minEdge;
      abox.uy = (randomSource.nextDouble() * edgeRange) + minEdge;
      aboxArea =  abox.ux * abox.uy;
      if (aboxArea + cumArea <= targetArea) {
        boxList.addElement(abox);
        cumArea += aboxArea;
      } else {
        break;
      }
    }
    double needArea = targetArea - cumArea;
    if (needArea <= 0) return;

    // have a little leftover area to attribute.
    if (boxList.size() == 0) {
      // OK...one square.
      abox = new Mcr();
      abox.lx = abox.ly = 0;
      abox.ux = abox.uy = Math.sqrt(needArea);
      return;
    }
    // normally, we have some boxes. One gets bigger. The first one.
    abox = (Mcr) boxList.elementAt(0);
    double oldarea = abox.ux * abox.uy;
    double newarea = oldarea + needArea;
    // make it square....
    abox.ux = Math.sqrt(newarea);
    abox.uy = Math.sqrt(newarea);
    // and it's already in the boxList, so just leave now
  }

  /** Sort by awkwardness to place--biggest dimension first. Bubblesort. */
  static private void sortBoxes() {
    for (int i=0; i<boxList.size() - 1; i++) {
      for (int j=i+1; j<boxList.size(); j++) {
        Mcr a = (Mcr)boxList.elementAt(i);
        Mcr b = (Mcr)boxList.elementAt(j);
        if (boxLT(a,b)) {
          boxList.setElementAt(b, i);
          boxList.setElementAt(a, j);
        }
      }
    }
  }
  /** box a "less than" box b, using only ux and uy */
  static private boolean boxLT(Mcr a, Mcr b) {
    return (a.ux < b.ux) && (a.ux < b.uy) &&
           (a.uy < b.ux) && (a.uy < b.uy);
  }

  static private void placeBoxes() {
    Quad root = new Quad(); // root of working tree

    int savenap = 4;
    // for debug purposes
    if (watch) {
      boolean firsttime = (watchFrame == null);
      if (firsttime) {
        watchFrame = new UnblinkFrame();
        watchFrame.setBounds(0,0, 400, 400);
        quadd = new QuadDisplay(root);
        quadd.setBounds(0,0, 400, 400);
        quadd.setBackground(Color.black);
        quadd.setVisible(true);
        watchFrame.add(quadd);
        watchFrame.pack();
        watchFrame.show();
        watchFrame.repaint();
      } else {
        quadd.setRoot(root);
        watchFrame.repaint();
      }
      MemCleaner.setNapTime(50);
      Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
    }
    // end debug enabling
    Quad collector = new Quad();
    collector.setType(Quad.QUAD_SESSILE);
    root.addChild(collector);
    SpaceQuad newbie;
    double dx, dy;
    for (int i=0; i<boxList.size(); i++) {
      if (watch)
        Globals.setCurrentTime(i);

      // initial random placement
      newbie = new SpaceQuad( (Mcr) boxList.elementAt(i), overArea, cover, i);
      dx = (overArea.getLenX() - newbie.ux) * randomSource.nextDouble();
      dy = (overArea.getLenY() - newbie.uy) * randomSource.nextDouble();
      newbie.moveCenter(overArea.lx + dx, overArea.ly + dy);
      collector.addNewbie(newbie);
      if (watch) {
        synchronized(watchFrame.getGraphics()) {
          watchFrame.repaint();
          Thread.currentThread().yield();
          MemCleaner.nap();
        }
      }

      // jostle-fest
      if (i != 0) {
        // sort the tree occassionally, if there are a lot of boxes
        if ((i > 20) && (i % 5 == 0))
          root.rootQuad();
        int jostleBailout = i * 4;
        if (jostleBailout < 10) jostleBailout = 10;
        jostle(root, jostleBailout);
      }

    }
    // that's it, they're placed. The SpaceQuad updates the boxList Mcr it came from.

    if (watch) {
      watchFrame.repaint();
      Globals.setCurrentTime(0);
      MemCleaner.setNapTime(savenap);
    }

  }

  /** returns true if jostle converged, false if maxLoco'd out */
  static boolean jostle(Quad root, int limit) {
    SpaceQuad curPusher;
    SpaceQuad curPushed;
    Mcr curPushWin;

    // give everyone a chance to push everyone else
    int moveCount = 1; // just get past while loop first time
    int locoCount = 0;
    while ((moveCount > 0) && (locoCount < limit)) {
      moveCount = 0;
      locoCount++;
      curPusher = (SpaceQuad) root.pickWin(root, root, true);

      while (curPusher != null) {
        curPushWin = curPusher.getSpaceMcr();
        curPushed = (SpaceQuad) curPusher.pickWin(curPushWin, root, true);
        while (curPushed != null) {
          curPusher.push(curPushed);
          curPushed = (SpaceQuad) curPushed.pickWin(curPushWin, root, true);
        }
        // at this point, I've gotten all the pushes I can get, since
        // the traversal is proceeding ahead of me.
        if (curPusher.applyDirection())
          moveCount++;
        curPusher = (SpaceQuad) curPusher.pickWin(root, root, true);
      }
      if (watch) {
        watchFrame.repaint();
        Thread.currentThread().yield();
        MemCleaner.nap();
      }
    }
    return (moveCount == 0);
  }
}

class SpaceQuad extends Quad {
  Mcr whoami;
  int id; // convenience for debug
  Mcr world;
  double cover;
  double cx, cy;
  double rx, ry;
  double sx, sy;
  Direction direction;
  Mcr spacem;
  Mcr crossm;

  /** The Mcr passed in has 0,0,ux,uy as dimensions.*/
  SpaceQuad(Mcr m, Mcr worldMcr, double coverPercent, int arbId) {
    super();
    id = arbId;
    type = QUAD_FEED;
    whoami = m;
    setFrom(m);
    world = worldMcr;
    cover = coverPercent;
    cx = m.getCenterX();
    cy = m.getCenterY();
    rx = m.ux / 2.0;
    ry = m.uy / 2.0;
    // factor is used in determining desired min spacing
    // want, on average, area/spacearea = cover, but that's too constrained
    // so solve for area/spacearea = sqrt(cover) and (1+factor) multiplier
    // for x and y dimensions (not r, but dimension=2*r), and solve a
    // quadratic equation to get the expression for factor.
    // for cover = .75, factor is .075; for .5, it's .19; for 1.2, it's -.0446.
    double factor = -1.0 + Math.sqrt( 1.0 / ( Math.sqrt( cover ) ) );
    sx = m.ux * (1 + factor) / 2.0;
    sy = m.uy * (1 + factor) / 2.0;

    direction = new Direction();
    direction.wannax = direction.wannay = 0.0;
    spacem = new Mcr();
    crossm = new Mcr();
  }

  Mcr getSpaceMcr() {
    spacem.lx = cx - sx; spacem.ly = cy - sy;
    spacem.ux = cx + sx; spacem.uy = cy + sy;
    return spacem;
  }

  void push(SpaceQuad other) {
    spacem = getSpaceMcr();
    if (spacem.intersectMcr(other.getSpaceMcr(), crossm) == Mcr.MCR_DISJOINT)
      return;
    // if they're not disjoint, fractional overlap in x and y is used as push on
    // self and other. crossm has the overlap.
// push from center of each other
//    double pushx = cx - other.cx;
//    double pushy = cy - other.cy;
    double pushx = cx - crossm.getCenterX();
    double pushy = cy - crossm.getCenterY();
    if ((pushx == 0) && (pushx == 0)) {// I am the cross section
      pushx = cx - other.getCenterX();
      pushy = cy - other.getCenterY();
      if ((pushx == 0) && (pushy == 0)) { // oy, we are coincident
        pushx = 1;
        pushy = 1;
      }
    }

    // cross-weight--the fatter pushes harder
    direction.addDirection(pushx, pushy, crossm.getArea());

    pushx = other.cx - crossm.getCenterX();
    pushy = other.cy - crossm.getCenterY();
    if ((pushx == 0) && (pushx == 0)) {// I am the cross section
      pushx = other.cx - cx;
      pushy = other.cy - cy;
      if ((pushx == 0) && (pushy == 0)) { // oy, we are coincident
        pushx = 1;
        pushy = 1;
      }
    }
    other.direction.addDirection(pushx, pushy, crossm.getArea());
  }

  // apply wannas AND snap into world. Return true if moved.
  boolean applyDirection() {
    if ((direction.wannax == 0) && (direction.wannay == 0))
      return false;

// try normalized wannas
direction.getDirection();
// impossible to justify or explain....
double pushlen = Math.sqrt(Math.abs(sx-rx) * Math.abs(sy-ry));
// too close to coverage of 1.0, the boxes lose the ability to move
if ((pushlen < (.03 * rx)) || (pushlen < (.03 * ry)))
  pushlen = .03 * .5 * (rx + ry);
direction.wannax *= pushlen;
direction.wannay *= pushlen;

    cx += direction.wannax; cy += direction.wannay;
    setMcr();
    direction.wannax = direction.wannay = 0;
    snapInto(world);
    cx = getCenterX();
    cy = getCenterY();
    growMcr();
    whoami.setFrom(this);
    return true;
  }

  void moveCenter(double byx, double byy) {
    cx += byx;
    cy += byy;
    setMcr();
    snapInto(world);
    cx = getCenterX();
    cy = getCenterY();
    growMcr();
    whoami.setFrom(this);
  }
  void setMcr() {
    lx = cx - rx;
    ly = cy - ry;
    ux = cx + rx;
    uy = cy + ry;
  }

  public void paint(QuadDisplay qd) {
    qd.drawRect(Color.green, cx-sx, cy-sy, cx+sx, cy+sy, false);
    qd.drawRect(Color.red, cx-rx, cy-ry, cx+rx, cy+ry, false);
  }
} /* class SpaceQuad */



