
//Title:        Indeco Steel Simulator
//Version:
//Copyright:    Copyright (c) 1999
//Author:       Joseph Guse & Ginger Booth
//Company:      Yale FES/CCE
//Description:  Steel industry economic simulator

/** SumTrace tracks a sum as a plottable statistic. That which is summed
 *  comes from a collection of StatBrokerable objects, which squeeze current
 *  values, weights, and normalization factors out of whoever's got the
 *  real data du jour. If somebody dies, it must be removed from the SumTrace.
 *
 *  SumTrace data is weighted and normalized on a per-stat-source-member basis.
 */
/** Adopted from Indeco for Gecko-in-CourseWare, 990708, vfb. */

package stats;

import java.util.*;
import java.awt.Color;
import java.lang.reflect.*;
import courseware.plot.*;

public class SumTrace extends DataTrace implements Steppable {
  protected Vector members;       // of StatBrokerable
  protected Vector membersX;       // of StatBrokerable
  private boolean xTimeFlag = false;
  protected final static int alloc = 50;
  public final static Color defaultColor = Color.black;

  public SumTrace(String title) {
    this(title, defaultColor, (Timer)null);
  }
  public SumTrace(String title, Color c) {
    this(title, c, (Timer)null);
  }
  public SumTrace() { super(alloc); }
  public SumTrace(String title, Color color, Timer timer) {
    super(alloc);
    setLabel(title);
    setColor(color);
    if (timer != null)
      setXTime(timer);
  }
  public SumTrace(String title, Color color, DoubleTimer timer) {
    super(alloc);
    setLabel(title);
    setColor(color);
    if (timer != null)
      setXTime(timer);
  }
  public void setXTime(Timer timer) {
    membersX = new Vector(1,0);
    membersX.addElement(new TimeBroker(timer));
    xTimeFlag = true;
  }
  public void setXTime(DoubleTimer timer) {
    membersX = new Vector(1,0);
    membersX.addElement(new DoubleTimeBroker(timer));
    xTimeFlag = true;
  }
  public boolean isXTime() { return xTimeFlag;}

  /** Add member as a statsource for this sum.  If you've added this StatBroker
   *  or source of a StatBroker before...OK. You're the boss.
   */
  public void addY(StatBrokerable broker)
  {
    if (members == null)
      members = new Vector(alloc, alloc);
    members.addElement(broker);

    class Deader implements StatSourceDeathListener {
      StatBrokerable theBroker;
      public Deader(StatBrokerable broker) {theBroker=broker;}
      public void onDeath(StatSourceDeathEvent ade){
        members.removeElement(theBroker);
      }
    }
    if (broker.getStatSource() instanceof StatSource) {
      ((StatSource)broker.getStatSource()).addStatSourceDeathListener( new Deader(broker) );
    }
  }
  /** If want the x-axis to be time, use setXTime instead. */
  public void addX(StatBrokerable broker)
  {
    if (membersX == null)
      membersX = new Vector(alloc, alloc);
    membersX.addElement(broker);

    class Deader implements StatSourceDeathListener {
      StatBrokerable theBroker;
      public Deader(StatBrokerable broker) {theBroker=broker;}
      public void onDeath(StatSourceDeathEvent ade){
        membersX.removeElement(theBroker);
      }
    }
    if (broker.getStatSource() instanceof StatSource) {
      ((StatSource)broker.getStatSource()).addStatSourceDeathListener( new Deader(broker) );
    }
    xTimeFlag = false;
  }

  /** The "source" we're looking for here is the one retured by "getStatSource"
   *  on a StatBrokerable. Most likely, the StatSource providing data values. The
   *  int returned is the relevant index into the members or membersX vector.
   */
  protected int find(Object source, Vector members) {
    int count = members.size();
    for (int i=0; i<count; i++) {
      if ( ( (StatBrokerable) members.elementAt(i) ).getStatSource() == source)
        return i;
    }
    return -1;
  }

  public int getMemberCount() {
    if (members == null) return 0;
    else return members.size();
  }
  public int getMemberXCount() {
    if (membersX == null) return 0;
    else return membersX.size();
  }

  /** Remove one member, only. Please use removeAll for all of them.... */
  public void remove(Object source) {
    int indexWas = find(source, members);
    if (indexWas != -1) {
      members.removeElementAt(indexWas);
      members.trimToSize();
    }
    indexWas = find(source, membersX);
    if (indexWas != -1) {
      membersX.removeElementAt(indexWas);
      membersX.trimToSize();
    }
  }

  public void removeAll() {members = null; membersX = null;}

  public static boolean invalid(double z) {
    Double tester = new Double(z);
    if (tester.isNaN() || tester.isInfinite())
      return true;
    else
      return false;
  }

  /** SumTrace simply ignores invalid members. */
  public double doSum(Vector members) {
    if (members == null)
      return 0.0;
    double sum = 0;
    int nvalid = 0;
    StatBrokerable broker;
    double value, weight;
    for (int i=0; i<members.size(); i++) {
      broker = (StatBrokerable) members.elementAt(i);
      if ( !( invalid( value=broker.getValue() ) ) ) {
        if ( !( invalid( weight=broker.getWeight() ) ) ) {
          sum += value * weight * broker.getNormalizer();
          nvalid++;
        }
      }
    }
    // need to distinguish between legitimate sum of 0.0 and
    // no-valid-values-found
    if (nvalid == 0)
      return StatBrokerable.INVALID_STAT;
    else
      return sum;
  }

  public void step() {
    super.add(doSum(membersX), doSum(members));
  }

  /** atIndex should basically be "index of now", zero at the beginning. */
  public void start() {
    super.setX(0, doSum(membersX));
    super.setX(0, doSum(members));
  }

  public String toString() {
    return "SumTrace " + getLabel();
  }
}
